# -*- coding: utf-8 -*-
#
# Extra tests which are not part of the SHT or DASH test suites,
# nor the discrete issues tests or the cmdline_test file.
# The need for these tests are discovered by doing coverage checks and these
# are added as required.
import os
import re

from rdflib import Graph

from pyshacl import validate
from pyshacl.errors import ReportableRuntimeError

ontology_file_text = """
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exOnt: <http://example.com/exOnt#> .

<http://example.com/exOnt> a owl:Ontology ;
    rdfs:label "An example extra-ontology file."@en .

exOnt:Animal a rdfs:Class ;
    rdfs:comment "The parent class for Humans and Pets"@en ;
    rdfs:subClassOf owl:Thing .

exOnt:Human a rdfs:Class ;
    rdfs:comment "A Human being"@en ;
    rdfs:subClassOf exOnt:Animal .

exOnt:Pet a rdfs:Class ;
    rdfs:comment "An animal owned by a human"@en ;
    rdfs:subClassOf exOnt:Animal .

exOnt:hasPet a rdf:Property ;
    rdfs:domain exOnt:Human ;
    rdfs:range exOnt:Pet .

exOnt:nLegs a rdf:Property ;
    rdfs:domain exOnt:Animal ;
    rdfs:range xsd:integer .

exOnt:Teacher a rdfs:Class ;
    rdfs:comment "A Human who is a teacher."@en ;
    rdfs:subClassOf exOnt:Human .

exOnt:PreschoolTeacher a rdfs:Class ;
    rdfs:comment "A Teacher who teaches preschool."@en ;
    rdfs:subClassOf exOnt:Teacher .

exOnt:Lizard a rdfs:Class ;
    rdfs:subClassOf exOnt:Pet .

exOnt:Goanna a rdfs:Class ;
    rdfs:subClassOf exOnt:Lizard .

"""

shacl_file_text = """
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exShape: <http://example.com/exShape#> .
@prefix exOnt: <http://example.com/exOnt#> .

<http://example.com/exShape> a owl:Ontology ;
    rdfs:label "Example Shapes File"@en .

exShape:HumanShape a sh:NodeShape ;
    sh:property [
        sh:class exOnt:Pet ;
        sh:path exOnt:hasPet ;
    ] ;
    sh:property [
        sh:datatype xsd:integer ;
        sh:path exOnt:nLegs ;
        sh:maxInclusive 2 ;
        sh:minInclusive 2 ;
    ] ;
    sh:targetClass exOnt:Human .

exShape:AnimalShape a sh:NodeShape ;
    sh:property [
        sh:datatype xsd:integer ;
        sh:path exOnt:nLegs ;
        sh:maxInclusive 4 ;
        sh:minInclusive 1 ;
    ] ;
    sh:targetClass exOnt:Animal .
"""

data_file_text = """
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exOnt: <http://example.com/exOnt#> .
@prefix ex: <http://example.com/ex#> .

ex:Human1 rdf:type exOnt:PreschoolTeacher ;
    rdf:label "Amy" ;
    exOnt:nLegs "2"^^xsd:integer ;
    exOnt:hasPet ex:Pet1 .

ex:Pet1 rdf:type exOnt:Goanna ;
    rdf:label "Sebastian" ;
    exOnt:nLegs "4"^^xsd:integer .
"""

data_file_text_bad = """
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exOnt: <http://example.com/exOnt#> .
@prefix ex: <http://example.com/ex#> .

ex:Human1 rdf:type exOnt:PreschoolTeacher ;
    rdf:label "Amy" ;
    exOnt:nLegs "2"^^xsd:integer ;
    exOnt:hasPet "Sebastian"^^xsd:string .

ex:Pet1 rdf:type exOnt:Goanna ;
    rdf:label "Sebastian" ;
    exOnt:nLegs "four"^^xsd:string .
"""


def test_validate_with_ontology():
    g = Graph().parse(data=data_file_text, format='turtle')
    e = Graph().parse(data=ontology_file_text, format='turtle')
    g_len = len(g)
    res = validate(
        g, shacl_graph=shacl_file_text, shacl_graph_format='turtle', ont_graph=e, inference='both', debug=True
    )
    conforms, graph, string = res
    g_len2 = len(g)
    assert conforms
    assert g_len2 == g_len


def test_validate_with_ontology_inplace():
    g = Graph().parse(data=data_file_text, format='turtle')
    e = Graph().parse(data=ontology_file_text, format='turtle')
    g_len = len(g)
    res = validate(
        g,
        shacl_graph=shacl_file_text,
        shacl_graph_format='turtle',
        ont_graph=e,
        inference='both',
        debug=True,
        inplace=True,
    )
    conforms, graph, string = res
    g_len2 = len(g)
    assert conforms
    assert g_len2 != g_len


def test_validate_with_ontology_fail1():
    res = validate(
        data_file_text_bad,
        shacl_graph=shacl_file_text,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        ont_graph=ontology_file_text,
        ont_graph_format="turtle",
        inference='both',
        debug=True,
    )
    conforms, graph, string = res
    assert not conforms


def test_validate_with_ontology_fail2():
    res = validate(
        data_file_text_bad,
        shacl_graph=shacl_file_text,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        ont_graph=ontology_file_text,
        ont_graph_format="turtle",
        inference=None,
        debug=True,
    )
    conforms, graph, string = res
    assert not conforms


def test_metashacl_pass():
    res = validate(
        data_file_text,
        shacl_graph=shacl_file_text,
        meta_shacl=True,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        ont_graph=ontology_file_text,
        ont_graph_format="turtle",
        inference='both',
        debug=True,
    )
    conforms, graph, string = res
    assert conforms


def test_metashacl_fail():
    bad_shacl_text = """
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix ex: <http://example.com/ex#> .

ex:HumanShape a sh:NodeShape ;
    sh:property [
        sh:class ex:Pet ;
        sh:path "2"^^xsd:integer ;
    ] ;
    sh:property [
        sh:datatype xsd:integer ;
        sh:path ex:nLegs ;
        sh:maxInclusive 2 ;
        sh:minInclusive 2 ;
    ] ;
    sh:targetClass ex:Human .

ex:AnimalShape a sh:NodeShape ;
    sh:property [
        sh:datatype xsd:integer ;
        sh:path ex:nLegs ;
        sh:maxInclusive 4 ;
        sh:minInclusive 1 ;
    ] ;
    sh:targetClass ex:Animal .
"""
    did_error = False
    try:
        res = validate(
            data_file_text,
            shacl_graph=bad_shacl_text,
            meta_shacl=True,
            data_graph_format='turtle',
            shacl_graph_format='turtle',
            ont_graph=ontology_file_text,
            ont_graph_format="turtle",
            inference='both',
            debug=True,
        )
        conforms, graph, string = res
        assert not conforms
    except ReportableRuntimeError as r:
        assert "Shapes SHACL (MetaSHACL) file." in r.message
        did_error = True
    assert did_error


data_file_text_bn = """
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exOnt: <http://example.com/exOnt#> .
@prefix ex: <http://example.com/ex#> .

ex:Student1 exOnt:hasTeacher [
    rdf:type exOnt:PreschoolTeacher ;
    rdf:label "Amy" ;
    exOnt:nLegs "2"^^xsd:integer ;
    exOnt:hasPet ex:Pet1 ]
.

ex:Pet1 rdf:type exOnt:Goanna ;
    rdf:label "Sebastian" ;
    exOnt:nLegs "4"^^xsd:integer .
"""

data_file_text_bad_bn = """
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exOnt: <http://example.com/exOnt#> .
@prefix ex: <http://example.com/ex#> .

ex:Student1 exOnt:hasTeacher [
    rdf:type exOnt:PreschoolTeacher ;
    rdf:label "Amy" ;
    exOnt:nLegs "2"^^xsd:integer ;
    exOnt:hasPet "Sebastian"^^xsd:string ]
.

ex:Pet1 rdf:type exOnt:Goanna ;
    rdf:label "Sebastian" ;
    exOnt:nLegs "four"^^xsd:string .
"""


def test_blank_node_string_generation():

    res = validate(
        data_file_text_bad_bn,
        shacl_graph=shacl_file_text,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        ont_graph=ontology_file_text,
        ont_graph_format="turtle",
        inference='rdfs',
        debug=True,
    )
    conforms, graph, string = res
    assert not conforms
    rx = r"^\s*Focus Node\:\s+\[.+rdf:type\s+exOnt\:PreschoolTeacher.*\]$"
    matches = re.search(rx, string, flags=re.MULTILINE)
    assert matches


def test_serialize_report_graph():
    res = validate(
        data_file_text,
        shacl_graph=shacl_file_text,
        data_graph_format='turtle',
        serialize_report_graph=True,
        shacl_graph_format='turtle',
        ont_graph=ontology_file_text,
        ont_graph_format="turtle",
        inference='both',
        debug=True,
    )
    conforms, graph, string = res
    assert isinstance(graph, (str, bytes))


shacl_file_property_shapes_text = """\
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exShape: <http://example.com/exShape#> .
@prefix exOnt: <http://example.com/exOnt#> .

<http://example.com/exShape> a owl:Ontology ;
    rdfs:label "Example Shapes File"@en .

exShape:HumanHasPetShape a sh:PropertyShape ;
    sh:class exOnt:Pet ;
    sh:path exOnt:hasPet ;
    sh:targetClass exOnt:Human .

exShape:HumanHasLegsShape a sh:PropertyShape ;
    sh:datatype xsd:integer ;
    sh:path exOnt:nLegs ;
    sh:maxInclusive 2 ;
    sh:minInclusive 2 ;
    sh:targetClass exOnt:Human .

exShape:PetHasLegsShape a sh:PropertyShape ;
    sh:datatype xsd:integer ;
    sh:path exOnt:nLegs ;
    sh:maxInclusive 4 ;
    sh:minInclusive 1 ;
    sh:targetClass exOnt:Animal .
"""


def test_property_shape_focus():
    res = validate(
        data_file_text,
        shacl_graph=shacl_file_property_shapes_text,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        ont_graph=ontology_file_text,
        ont_graph_format="turtle",
        inference='rdfs',
        debug=True,
    )
    conforms, graph, string = res
    assert conforms


def test_property_shape_focus_fail1():
    res = validate(
        data_file_text_bad,
        shacl_graph=shacl_file_property_shapes_text,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        ont_graph=ontology_file_text,
        ont_graph_format="turtle",
        inference='rdfs',
        debug=True,
    )
    conforms, graph, string = res
    assert not conforms


web_d1_ttl = """\
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exOnt: <http://example.com/exOnt#> .
@prefix ex: <http://example.com/ex#> .

ex:Human1 rdf:type exOnt:Human ;
    rdf:label "Amy" ;
    exOnt:nLegs "2"^^xsd:integer ;
    exOnt:hasPet ex:Pet1 .

ex:Pet1 rdf:type exOnt:Lizard ;
    rdf:label "Sebastian" ;
    exOnt:nLegs "4"^^xsd:integer .
"""
web_d2_ttl = """\
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exOnt: <http://example.com/exOnt#> .
@prefix ex: <http://example.com/ex#> .

ex:Human1 rdf:type exOnt:Human ;
    rdf:label "Amy" ;
    exOnt:nLegs "2"^^xsd:integer ;
    exOnt:hasPet "Sebastian"^^xsd:string .

ex:Pet1 rdf:type exOnt:Lizard ;
    rdf:label "Sebastian" ;
    exOnt:nLegs "g"^^xsd:string .
"""


def test_web_retrieve():
    DEB_BUILD_ARCH = os.environ.get('DEB_BUILD_ARCH', None)
    DEB_HOST_ARCH = os.environ.get('DEB_HOST_ARCH', None)
    if DEB_BUILD_ARCH is not None or DEB_HOST_ARCH is not None:
        print("Cannot run web requests in debhelper tests.")
        assert True
        return True
    shacl_file = "https://raw.githubusercontent.com/RDFLib/pySHACL/master/test/resources/cmdline_tests/s1.ttl"
    ont_file = "https://raw.githubusercontent.com/RDFLib/pySHACL/master/test/resources/cmdline_tests/o1.ttl"
    res = validate(
        web_d1_ttl,
        shacl_graph=shacl_file,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        ont_graph=ont_file,
        ont_graph_format="turtle",
        inference='both',
        debug=True,
    )
    conforms, graph, string = res
    assert conforms


def test_web_retrieve_fail():
    DEB_BUILD_ARCH = os.environ.get('DEB_BUILD_ARCH', None)
    DEB_HOST_ARCH = os.environ.get('DEB_HOST_ARCH', None)
    if DEB_BUILD_ARCH is not None or DEB_HOST_ARCH is not None:
        print("Cannot run web requests in debhelper tests.")
        assert True
        return True
    shacl_file = "https://raw.githubusercontent.com/RDFLib/pySHACL/master/test/resources/cmdline_tests/s1.ttl"
    ont_file = "https://raw.githubusercontent.com/RDFLib/pySHACL/master/test/resources/cmdline_tests/o1.ttl"
    res = validate(
        web_d2_ttl,
        shacl_graph=shacl_file,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        ont_graph=ont_file,
        ont_graph_format="turtle",
        inference='both',
        debug=True,
    )
    conforms, graph, string = res
    assert not conforms


my_partial_shapes_text = """
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix ex: <http://example.com/ex1#> .

<http://example.com/ex1> a owl:Ontology ;
    owl:imports <https://raw.githubusercontent.com/RDFLib/pySHACL/master/test/resources/cmdline_tests/s1.ttl> .
"""

my_partial_ont_text = """
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix ex: <http://example.com/ex2#> .

<http://example.com/ex2> a owl:Ontology ;
    owl:imports <https://raw.githubusercontent.com/RDFLib/pySHACL/master/test/resources/cmdline_tests/o1.ttl> .
"""


def test_owl_imports():
    DEB_BUILD_ARCH = os.environ.get('DEB_BUILD_ARCH', None)
    DEB_HOST_ARCH = os.environ.get('DEB_HOST_ARCH', None)
    if DEB_BUILD_ARCH is not None or DEB_HOST_ARCH is not None:
        print("Cannot run owl:imports in debhelper tests.")
        assert True
        return True
    res = validate(
        web_d1_ttl,
        shacl_graph=my_partial_shapes_text,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        ont_graph=my_partial_ont_text,
        ont_graph_format="turtle",
        inference='both',
        debug=True,
        do_owl_imports=True,
    )
    conforms, graph, string = res
    print(string)
    assert conforms


def test_owl_imports_fail():
    DEB_BUILD_ARCH = os.environ.get('DEB_BUILD_ARCH', None)
    DEB_HOST_ARCH = os.environ.get('DEB_HOST_ARCH', None)
    if DEB_BUILD_ARCH is not None or DEB_HOST_ARCH is not None:
        print("Cannot run owl:imports in debhelper tests.")
        assert True
        return True

    res = validate(
        web_d2_ttl,
        shacl_graph=my_partial_shapes_text,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        ont_graph=my_partial_ont_text,
        ont_graph_format=None,
        inference='both',
        debug=True,
        do_owl_imports=True,
    )
    conforms, graph, string = res
    print(string)
    assert not conforms


def test_sparql_message_subst():
    df = '''@prefix ex: <http://datashapes.org/sh/tests/#> .
    @prefix owl: <http://www.w3.org/2002/07/owl#> .
    @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
    @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
    @prefix sh: <http://www.w3.org/ns/shacl#> .
    @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

    ex:ValidResource1
      rdf:type rdfs:Resource ;
    .
    ex:InvalidResource1
      rdf:type rdfs:Resource ;
      rdfs:label "Invalid resource 1" ;
    .
    ex:InvalidResource2
      rdf:type rdfs:Resource ;
      rdfs:label "Invalid label 1" ;
      rdfs:label "Invalid label 2" ;
    .
    ex:TestShape
      rdf:type sh:NodeShape ;
      rdfs:label "Test shape" ;
      sh:sparql ex:TestShape-sparql ;
      sh:targetNode ex:InvalidResource1 ;
      sh:targetNode ex:InvalidResource2 ;
      sh:targetNode ex:ValidResource1 ;
    .
    ex:TestShape-sparql
      sh:message "{$this} cannot have a {$path} of {$value}" ;
      sh:prefixes <http://datashapes.org/sh/tests/sparql/node/sparql-001.test> ;
      sh:select """
        SELECT $this ?path ?value
        WHERE {
            $this ?path ?value .
            FILTER (?path = <http://www.w3.org/2000/01/rdf-schema#label>) .
        }""" ;
    .'''
    res = validate(
        df,
        data_graph_format='turtle',
        inference=None,
        debug=True,
    )
    conforms, graph, s = res
    assert "#InvalidResource1 cannot have a http://www.w3.org/2000/01/rdf-schema#label of Invalid resource 1" in s
    assert "#InvalidResource2 cannot have a http://www.w3.org/2000/01/rdf-schema#label of Invalid label 1" in s
    assert "#InvalidResource2 cannot have a http://www.w3.org/2000/01/rdf-schema#label of Invalid label 2" in s
    assert not conforms


if __name__ == "__main__":
    test_validate_with_ontology()
    test_validate_with_ontology_fail1()
    test_validate_with_ontology_fail2()
    test_metashacl_pass()
    test_metashacl_fail()
    test_blank_node_string_generation()
    test_property_shape_focus()
    test_property_shape_focus_fail1()
    test_web_retrieve()
    test_serialize_report_graph()
    test_owl_imports()
    test_owl_imports_fail()
    test_sparql_message_subst()



# -*- coding: utf-8 -*-
#
import sys
from functools import lru_cache
from warnings import warn

if sys.version_info[:2] < (3, 12):
    from importlib_metadata import PackageNotFoundError, metadata
else:
    from importlib.metadata import PackageNotFoundError, metadata


# In dev mode, the extras-loader doesn't check if extension is installed before loading it.
# This is useful when testing, when we're working on a codebase of a library that is not yet installed.
dev_mode = False


extras_requirements = {"js": ["pyduktape2"], "http": ["sanic", "sanic-ext", "sanic-cors"]}


@lru_cache()
def check_extra_installed(extra_name: str):
    if dev_mode:
        return True
    # first check if pyshacl is installed using the normal means
    try:
        this_mdata = metadata('pyshacl')
    except PackageNotFoundError:
        # Hmm, it thinks pyshacl isn't installed. Can't even check for extras
        return None
    try:
        has_extras = this_mdata.json['provides_extra']
    except LookupError:
        # Can't check metadata for extras for some reason?
        return None
    if extra_name not in has_extras:
        warn(Warning(f"Extra \"{extra_name}\" doesn't exist in this version of pyshacl."))
        return False
    if extra_name not in extras_requirements:
        warn(Warning(f"Extra \"{extra_name}\" cannot be checked in this version of pyshacl."))
    all_reqs = extras_requirements[extra_name]
    for req in all_reqs:
        try:
            _ = metadata(req)
        except PackageNotFoundError:
            warn(Warning(f"Extra \"{extra_name}\" is not satisfied because requirement {req} is not installed."))
            return False
    return True



# -*- coding: utf-8 -*-
#
"""
https://github.com/RDFLib/pySHACL/issues/281
"""
import sys

import rdflib

import pyshacl

# Tests are always run from project root, even nested issues tests.
BASE_DIR = "./test/issues/test_281"

def test_281_a():
    conforms, results_graph, results_text = pyshacl.validate(
        f"file:{BASE_DIR}/data.ttl",
        shacl_graph=f"file:{BASE_DIR}/policies.ttl",
        debug=True,
        inference='none',
        advanced=False,
        meta_shacl=False,
        do_owl_imports=True,
    )
    assert not conforms

def test_281_b():
    from pyshacl.rdfutil.load import load_from_source
    with open(f"{BASE_DIR}/data.ttl") as data_f:
        data_g = load_from_source(data_f, rdf_format="turtle", do_owl_imports=False)
    with open(f"{BASE_DIR}/policies.ttl") as policy_f:
        policy_g = load_from_source(policy_f, rdf_format="turtle", do_owl_imports=True)

    conforms, results_graph, results_text = pyshacl.validate(
        data_g,
        shacl_graph=policy_g,
        debug=True,
        inference='none',
        advanced=False,
        meta_shacl=False,
        do_owl_imports=True,
    )
    assert not conforms


if __name__ == "__main__":
    test_281_a()
    sys.exit(test_281_b())



# -*- coding: utf-8 -*-
from decimal import Decimal
from typing import TYPE_CHECKING, Optional, Sequence

from rdflib import RDF, Literal

from pyshacl.consts import SH_condition, SH_deactivated, SH_order
from pyshacl.errors import RuleLoadError
from pyshacl.pytypes import RDFNode, SHACLExecutor

if TYPE_CHECKING:
    from rdflib.term import URIRef

    from pyshacl.pytypes import GraphLike

RDF_first = RDF.first


class SHACLRuleCondition(object):
    __slots__ = ("rule", "cond_shape")

    def __init__(self, rule, cond_shape):
        self.rule = rule
        self.cond_shape = cond_shape

    def get_focus_nodes(self, data_graph):
        return self.cond_shape.focus_nodes(data_graph)

    def validate_condition(self, executor, data_graph, focus_node):
        return self.cond_shape.validate(executor, data_graph, focus=focus_node)


class SHACLRule(object):
    __slots__ = ("executor", "shape", "node", "iterate", "_deactivated")

    def __init__(self, executor: SHACLExecutor, shape, rule_node, iterate=False):
        """
        :param executor:
        :type executor: SHACLExecutor
        :param shape:
        :type shape: Shape
        :param rule_node:
        :type rule_node: rdflib.Identifier
        """
        super(SHACLRule, self).__init__()
        self.executor = executor
        self.shape = shape
        self.node = rule_node
        self.iterate = iterate

        deactivated_nodes = list(self.shape.sg.objects(self.node, SH_deactivated))
        self._deactivated = len(deactivated_nodes) > 0 and bool(deactivated_nodes[0])

    @property
    def deactivated(self):
        return self._deactivated

    @property
    def order(self) -> Decimal:
        order_nodes = list(self.shape.sg.objects(self.node, SH_order))
        if len(order_nodes) < 1:
            return Decimal("0.0")
        if len(order_nodes) > 1:
            raise RuleLoadError(
                "A SHACL Rule can have only one sh:order property.", "https://www.w3.org/TR/shacl-af/#rules-order"
            )
        order_node = next(iter(order_nodes))
        if not isinstance(order_node, Literal):
            raise RuleLoadError(
                "A SHACL Rule must be a numeric literal.", "https://www.w3.org/TR/shacl-af/#rules-order"
            )
        return Decimal(order_node.value)

    def get_conditions(self):
        shapes_graph_g = self.shape.sg.graph
        cond_nodes = list(shapes_graph_g.objects(self.node, SH_condition))
        conditions = []
        for c in cond_nodes:
            # test_me = list(self.shape.sg.graph.predicate_objects(c))
            # check if this is a rdf:list
            first_nodes = list(shapes_graph_g.objects(c, RDF_first))
            if len(first_nodes) > 0:
                for c_item in shapes_graph_g.items(c):
                    try:
                        cond_shape = self.shape.sg.lookup_shape_from_node(c_item)
                    except (AttributeError, KeyError):
                        raise RuleLoadError(
                            "A SHACL Rule Condition must be an existing well-formed SHACL Shape.",
                            "https://www.w3.org/TR/shacl-af/#condition",
                        )
                    condition = SHACLRuleCondition(self, cond_shape)
                    conditions.append(condition)
            else:
                try:
                    cond_shape = self.shape.sg.lookup_shape_from_node(c)
                except (AttributeError, KeyError):
                    raise RuleLoadError(
                        "A SHACL Rule Condition must be an existing well-formed SHACL Shape.",
                        "https://www.w3.org/TR/shacl-af/#condition",
                    )
                condition = SHACLRuleCondition(self, cond_shape)
                conditions.append(condition)
        return conditions

    def filter_conditions(self, focus_nodes: Sequence[RDFNode], data_graph):
        conditions = self.get_conditions()
        applicable_focus_nodes = []
        for f in focus_nodes:
            not_applicable = False
            for c in conditions:
                _conforms, _reports = c.validate_condition(self.executor, data_graph, f)
                not_applicable = not_applicable or not (_conforms)
            if not not_applicable:
                applicable_focus_nodes.append(f)
        return applicable_focus_nodes

    def apply(
        self,
        data_graph: 'GraphLike',
        focus_nodes: Optional[Sequence[RDFNode]] = None,
        target_graph_identifier: Optional['URIRef'] = None,
    ):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
#
"""
https://github.com/RDFLib/pySHACL/issues/71
"""
import rdflib

from pyshacl import validate


shacl_file_text = """\
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix dc: <http://purl.org/dc/elements/1.1/> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xml: <http://www.w3.org/XML/1998/namespace> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix geo: <http://example.org/geo/> .
@prefix ex: <http://example.com#> .
@prefix : <http://example.com/issue/071#> .


<http://example.com/issue/071>
  rdf:type owl:Ontology .

:PlaceShape
    a sh:NodeShape ;
    sh:targetClass ex:Place ;
    sh:property [
        sh:path geo:asWKT ;
        sh:nodeKind sh:Literal ;
        sh:datatype rdfs:Literal ;
    ] ;
    sh:property [
        sh:path geo:asWKT ;
        sh:nodeKind sh:Literal ;
        sh:datatype rdfs:Datatype ;
    ] ;
    sh:property [
        sh:path geo:asWKT ;
        sh:nodeKind sh:Literal ;
        sh:datatype geo:wktLiteral ;
    ] ;
.
"""

data_file_text = """\
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix dc: <http://purl.org/dc/elements/1.1/> .
@prefix dct: <http://purl.org/dc/terms/> .
@prefix xml: <http://www.w3.org/XML/1998/namespace> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix geo: <http://example.org/geo/> .
@prefix ex: <http://example.com#> .

geo:hasSerialization
    a rdfs:Property ;
    a owl:DatatypeProperty ;
    rdfs:domain geo:Geometry ;
    rdfs:range rdfs:Literal ;
.

geo:asWKT
    a rdfs:Property ;
    a owl:DatatypeProperty ;
    rdfs:subPropertyOf geo:hasSerialization ;
    rdfs:domain geo:Geometry ;
    rdfs:range geo:wktLiteral ;
.

geo:wktLiteral
    a rdfs:Datatype ;
    dc:description "A Well-known Text serialization of a geometry object." ;
.

ex:Place
    a rdfs:Class ;
    rdfs:subClassOf geo:Geometry ;
.


ex:test1
    a ex:Place ;
    geo:asWKT "POINT(45.75 4.85)"^^geo:wktLiteral ;
    dc:title "test1" ;
.

"""


def test_071_positive():
    data = rdflib.Graph()
    data.parse(data=data_file_text, format="turtle")
    res = validate(
        data,
        shacl_graph=shacl_file_text,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        inference='rdfs',
        debug=True,
    )
    conforms, graph, string = res
    assert conforms


if __name__ == "__main__":
    test_071_positive()



# -*- coding: utf-8 -*-
#
# Extra tests which are not part of the SHT or DASH test suites,
# nor the discrete issues tests or the cmdline_test file.
# The need for these tests are discovered by doing coverage checks and these
# are added as required.


from pyshacl import validate

ontology_file_text = """
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exOnt: <http://example.com/exOnt#> .

<http://example.com/exOnt> a owl:Ontology ;
    rdfs:label "An example extra-ontology file."@en .

exOnt:Animal a rdfs:Class ;
    rdfs:comment "The parent class for Humans and Pets"@en ;
    rdfs:subClassOf owl:Thing .

exOnt:Human a rdfs:Class ;
    rdfs:comment "A Human being"@en ;
    rdfs:subClassOf exOnt:Animal .

exOnt:Pet a rdfs:Class ;
    rdfs:comment "An animal owned by a human"@en ;
    rdfs:subClassOf exOnt:Animal .

exOnt:hasPet a rdf:Property ;
    rdfs:domain exOnt:Human ;
    rdfs:range exOnt:Pet .

exOnt:nLegs a rdf:Property ;
    rdfs:domain exOnt:Animal ;
    rdfs:range xsd:integer .

exOnt:Teacher a rdfs:Class ;
    rdfs:comment "A Human who is a teacher."@en ;
    rdfs:subClassOf exOnt:Human .

exOnt:PreschoolTeacher a rdfs:Class ;
    rdfs:comment "A Teacher who teaches preschool."@en ;
    rdfs:subClassOf exOnt:Teacher .

exOnt:Lizard a rdfs:Class ;
    rdfs:subClassOf exOnt:Pet .

exOnt:Turtle a rdfs:Class ;
    rdfs:subClassOf exOnt:Pet .

exOnt:Goanna a rdfs:Class ;
    rdfs:subClassOf exOnt:Lizard .

"""

shacl_file_text = """
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exShape: <http://example.com/exShape#> .
@prefix exOnt: <http://example.com/exOnt#> .

<http://example.com/exShape> a owl:Ontology ;
    rdfs:label "Example Shapes File"@en .

exShape:HumanShape a sh:NodeShape ;
    sh:property [
        sh:class exOnt:Pet ;
        sh:path exOnt:hasPet ;
    ] ;
    sh:property [
        sh:datatype xsd:integer ;
        sh:path exOnt:nLegs ;
        sh:maxInclusive 2 ;
        sh:minInclusive 2 ;
    ] ;
    sh:targetClass exOnt:Human .

exShape:AnimalShape a sh:NodeShape ;
    sh:property [
        sh:datatype xsd:integer ;
        sh:path exOnt:nLegs ;
        sh:maxInclusive 4 ;
        sh:minInclusive 1 ;
    ] ;
    sh:targetClass exOnt:Animal .
"""

data_file_text = """
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exOnt: <http://example.com/exOnt#> .
@prefix ex: <http://example.com/ex#> .

ex:Human1 rdf:type exOnt:PreschoolTeacher ;
    rdf:label "Amy" ;
    exOnt:nLegs "2"^^xsd:integer ;
    exOnt:hasPet ex:Pet1 .

ex:Pet1 rdf:type exOnt:Goanna ;
    rdf:label "Sebastian" ;
    exOnt:nLegs "4"^^xsd:integer .

ex:Human2 rdf:type exOnt:PreschoolTeacher ;
    rdf:label "JoAnne" ;
    exOnt:nLegs "2"^^xsd:integer ;
    exOnt:hasPet ex:Pet2 .

ex:Pet2 rdf:type exOnt:Turtle ;
    rdf:label "Terrance" ;
    exOnt:nLegs "4"^^xsd:integer .

"""

data_file_text_bad = """
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exOnt: <http://example.com/exOnt#> .
@prefix ex: <http://example.com/ex#> .

ex:Human1 rdf:type exOnt:PreschoolTeacher ;
    rdf:label "Amy" ;
    exOnt:nLegs "2"^^xsd:integer ;
    exOnt:hasPet "Sebastian"^^xsd:string .

ex:Pet1 rdf:type exOnt:Goanna ;
    rdf:label "Sebastian" ;
    exOnt:nLegs "four"^^xsd:string .

ex:Human2 rdf:type exOnt:PreschoolTeacher ;
    rdf:label "JoAnne" ;
    exOnt:nLegs "2"^^xsd:integer ;
    exOnt:hasPet "Terrance"^^xsd:string .

ex:Pet2 rdf:type exOnt:Turtle ;
    rdf:label "Terrance" ;
    exOnt:nLegs "four"^^xsd:string .

"""


def test_validate_pass_manual_targeting_focus():
    res = validate(
        data_file_text,
        shacl_graph=shacl_file_text,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        ont_graph=ontology_file_text,
        ont_graph_format="turtle",
        inference='both',
        focus_nodes=["ex:Human1"],
        debug=True,
    )
    conforms, graph, string = res
    assert "Results (1)" not in string
    assert conforms


def test_validate_fail_manual_targeting_focus():
    res = validate(
        data_file_text_bad,
        shacl_graph=shacl_file_text,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        ont_graph=ontology_file_text,
        ont_graph_format="turtle",
        inference='both',
        focus_nodes=["ex:Human1"],
        debug=True,
    )
    conforms, graph, string = res
    assert "Results (1)" in string
    assert not conforms


def test_validate_fail_manual_targeting_shape():
    res = validate(
        data_file_text_bad,
        shacl_graph=shacl_file_text,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        ont_graph=ontology_file_text,
        ont_graph_format="turtle",
        inference='both',
        use_shapes=["exShape:HumanShape"],
        debug=True,
    )
    conforms, graph, string = res
    assert "Results (2)" in string
    assert not conforms


def test_validate_fail_manual_targeting_focus_with_shape():
    res = validate(
        data_file_text_bad,
        shacl_graph=shacl_file_text,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        ont_graph=ontology_file_text,
        ont_graph_format="turtle",
        inference='both',
        focus_nodes=["ex:Human1"],
        use_shapes=["exShape:HumanShape"],
        debug=True,
    )
    conforms, graph, string = res
    assert "Results (1)" in string
    assert not conforms



# -*- coding: utf-8 -*-
#
"""
https://github.com/RDFLib/pySHACL/issues/38

I think this is duplicate of test test_029.py
"""
from rdflib import Graph
from pyshacl import validate

shapes = Graph()
shapes.parse(
    data="""
    @prefix sh: <http://www.w3.org/ns/shacl#> .
    @prefix owl: <http://www.w3.org/2002/07/owl#> .
    @prefix ex: <http://example.org/ns#> .

    ex:Person
          a owl:Class ;
          a sh:NodeShape ;
          sh:property ex:NameConstraint ;
    .

    ex:NameConstraint
          a sh:PropertyShape ;
          sh:path ex:name ;
          sh:minCount 1 ;
        .
""",
    format="ttl",
)

data = Graph()
data.parse(
    data="""
    @prefix ex: <http://example.org/ns#> .

    ex:Bob
          a ex:Person ;
    .
""",
    format="ttl",
)


def test_038():
    conforms, g, s = validate(data_graph=data, shacl_graph=shapes, inference='rdfs')
    assert not conforms


if __name__ == "__main__":
    test_038()



# -*- coding: utf-8 -*-
#
"""
https://github.com/RDFLib/pySHACL/issues/76
"""
import rdflib

from pyshacl import validate


shacl_file_text = '''
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix skos: <http://www.w3.org/2004/02/skos/core#> .
@prefix xml: <http://www.w3.org/XML/1998/namespace> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix : <http://example.com/issue/076#> .

<http://example.com/issue/076>
  rdf:type owl:Ontology ;
  owl:sameAs : .

<http://example.com/issue/076#>
  rdf:type owl:Ontology ;
  owl:imports <http://datashapes.org/dash> ;
  sh:declare [
    sh:prefix "skos_p" ;
    sh:namespace "http://www.w3.org/2004/02/skos/core#"^^xsd:anyURI ;
    ] ;
  sh:declare [
    sh:prefix "" ;
    sh:namespace "http://example.com/issue/076#"^^xsd:anyURI ;
    ] .

:TopConceptRule
	a sh:NodeShape ;
	sh:property [
		sh:path skos:topConceptOf ;
		sh:minCount 1 ;
	] .

:DepthRule
	a sh:NodeShape ;
	sh:targetClass skos:Concept ;
	sh:rule [
		a sh:SPARQLRule ;
		sh:prefixes : ;
		sh:order 1 ;
		sh:condition :TopConceptRule ;
		sh:construct """
		    CONSTRUCT {
				$this :hasDepth 0 .
			}
			WHERE {
			}
		""" ;
	] ;
	sh:rule [
		a sh:SPARQLRule ;
		sh:prefixes : ;
		sh:order 2 ;
		sh:construct """
		    CONSTRUCT {
				$this :hasDepth ?plusOne .
			}
			WHERE {
				$this skos_p:broader ?parent .
				?parent :hasDepth ?depth .
				bind(?depth + 1 as ?plusOne)
			}
		""" ;
	] .
'''

data_file_text = """
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix skos: <http://www.w3.org/2004/02/skos/core#> .
@prefix dct: <http://purl.org/dc/terms/> .
@prefix xml: <http://www.w3.org/XML/1998/namespace> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix ex: <http://example.com#> .

ex:animalsVocabulary rdf:type skos:ConceptScheme;
  dct:title "Animals Vocabulary"@en;
  skos:hasTopConcept ex:animals .

ex:animals rdf:type skos:Concept;
  skos:prefLabel "animals"@en;
  skos:inScheme ex:animalsVocabulary;
  skos:topConceptOf ex:animalsVocabulary .

ex:cat rdf:type skos:Concept;
  skos:prefLabel "cat"@en;
  skos:broader ex:animals ;
  skos:inScheme ex:animalsVocabulary.

ex:wildcat a skos:Concept;
  skos:inScheme ex:animalsVocabulary;
  skos:broader ex:cat .

ex:europeanWildcat a skos:Concept;
  skos:inScheme ex:animalsVocabulary;
  skos:broader ex:wildcat .
"""


def test_076_positive():
    data = rdflib.Graph()
    data.parse(data=data_file_text, format="turtle")
    res = validate(
        data,
        shacl_graph=shacl_file_text,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        inference='rdfs',
        inplace=True,
        advanced=True,
        iterate_rules=True,
        debug=True,
    )
    conforms, graph, string = res
    find_s = rdflib.URIRef("http://example.com#europeanWildcat")
    find_p = rdflib.URIRef("http://example.com/issue/076#hasDepth")
    find_o = rdflib.Literal(3)
    assert (find_s, find_p, find_o) in data


def test_076_negative():
    data = rdflib.Graph()
    data.parse(data=data_file_text, format="turtle")
    res = validate(
        data,
        shacl_graph=shacl_file_text,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        inference='rdfs',
        inplace=True,
        advanced=True,
        iterate_rules=False,
        debug=True,
    )
    conforms, graph, string = res
    find_s = rdflib.URIRef("http://example.com#europeanWildcat")
    find_p = rdflib.URIRef("http://example.com/issue/076#hasDepth")
    find_o = rdflib.Literal(3)
    assert (find_s, find_p, find_o) not in data


if __name__ == "__main__":
    test_076_positive()
    test_076_negative()



from rdflib import Graph

from pyshacl import validate


"""
https://github.com/RDFLib/pySHACL/issues/70
"""

data_text = """\
@prefix ns0: <http://www.opengis.net/ont/geosparql#> .
@prefix dc: <http://purl.org/dc/terms/> .
@prefix ns1: <http://qudt.org/schema/qudt/> .
@prefix sosa: <http://www.w3.org/ns/sosa/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix rdfs:   <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs:  <http://www.w3.org/2000/01/rdf-schema#> .
@prefix geo:   <http://www.opengis.net/ont/geosparql#> .
@prefix sf:    <http://www.opengis.net/ont/sf#> .

<http://www.w3id.org/afarcloud/pCoord?lat=45.75&amp;long=4.85>
  ns0:asWKT "POINT(45.75 4.85)"^^ns0:wktLiteral ;
  a <http://www.opengis.net/ont/sf#Point> .

<http://www.w3id.org/afarcloud/poi?lat=45.75&amp;long=4.85>
  ns0:hasGeometry <http://www.w3id.org/afarcloud/pCoord?lat=45.75&amp;long=4.85> ;
  a ns0:Feature .

<urn:afc:AS09:cropsManagement:TEC:soil:sen0022> a <https://json-ld.org/playground/AfarcloudSensors>, <https://json-ld.org/playground/SoilSensor> .
<urn:afc:AS09:sen0022:obs-1514810172/q1>
  dc:identifier "q1" ;
  ns1:numericValue "0.27121272683143616" ;
  ns1:unit <http://qudt.org/vocab/unit/DEG_C> ;
  a ns1:QuantityValue .

<urn:afc:AS09:sen0022:obs-1514810172>
  a sosa:Observation ;
  sosa:hasFeatureOfInterest <http://www.w3id.org/afarcloud/poi?lat=45.75&amp;long=4.85> ;
  sosa:hasResult <urn:afc:AS09:sen0022:obs-1514810172/q1> ;
  sosa:madeBySensor <urn:afc:AS09:cropsManagement:TEC:soil:sen0022> ;
  sosa:observedProperty <http://www.w3id.org/afarcloud/soil_temperature> ;
  sosa:resultTime "2018-01-01T12:36:12Z"^^xsd:dateTime .
"""
shacl_url = """https://raw.githubusercontent.com/rapw3k/DEMETER/master/models/SHACL/demeterAgriProfile-SHACL.ttl"""


def test_070():
    g = Graph().parse(data=data_text, format="turtle")
    try:
        conforms, g, s = validate(
            g,
            shacl_graph=shacl_url,
            shacl_graph_format="turtle",
            abort_on_error=False,
            meta_shacl=False,
            debug=True,
            advanced=True,
        )
    except Exception as e:
        print(e)
        raise
    assert conforms


if __name__ == "__main__":
    test_070()



# -*- coding: utf-8 -*-
#

# Portions of this file contributed by NIST are governed by the following
# statement:
#
# This software was developed at the National Institute of Standards
# and Technology by employees of the Federal Government in the course
# of their official duties. Pursuant to Title 17 Section 105 of the
# United States Code, this software is not subject to copyright
# protection within the United States. NIST assumes no responsibility
# whatsoever for its use by other parties, and makes no guarantees,
# expressed or implied, about its quality, reliability, or any other
# characteristic.
#
# We would appreciate acknowledgement if the software is used.

"""
https://github.com/RDFLib/pySHACL/issues/160

This test demonstrates two styles of using a shape focused on a
Predicate to review the class of nodes that are Subjects for that
Predicate, using the sh:class constraint.

shacl_file_1_text uses a NodeShape and sh:targetSubjectsOf.

shacl_file_2_text uses a PropertyShape and an sh:inversePath.

The motivation for these tests was to clarify how to use sh:inversePath
and a sh:targetXOf selector; sh:targetSubjectsOf turns out to lead to
incorrect behavior.

Note that depending on the style used, sh:ValidationResults will link
the "focus" node as either sh:focusNode, or sh:value.
"""

from typing import Set

from rdflib import Graph, SH, URIRef

from pyshacl import validate


data_ontology_file_text = """\
@prefix ex: <http://example.org/ontology/> .
@prefix kb: <http://example.org/kb/> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

<http://example.org/kb>
  rdf:type owl:Ontology ;
  rdfs:label "Test of inversePath" ;
.
kb:thing-a-1
  rdf:type ex:ThingA ;
  ex:propertyOfA "1" ;
.
kb:thing-b-1
  rdf:type ex:ThingB ;
  ex:propertyOfA "1" ;
.
"""

shacl_file_1_text = """\
@prefix ex: <http://example.org/ontology/> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

ex:propertyOfA-nodeshape
  rdf:type sh:NodeShape ;
  sh:class ex:ThingA ;
  sh:targetSubjectsOf ex:propertyOfA ;
  .
"""

shacl_file_2_text = """\
@prefix ex: <http://example.org/ontology/> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

ex:propertyOfA-propertyshape
  rdf:type sh:PropertyShape ;
  sh:class ex:ThingA ;
  sh:path [
    sh:inversePath ex:propertyOfA ;
  ] ;
  sh:targetObjectsOf ex:propertyOfA ;
  .
"""


def _test_160_template(
    shacl_file_text: str,
    n_validation_result_property: URIRef,
) -> None:
    (
        conforms,
        conformance_graph,
        conformance_text,
    ) = validate(
        data_ontology_file_text,
        shacl_graph=shacl_file_text,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        debug=True,
        metashacl=True,
    )

    assert not conforms
    assert isinstance(conformance_graph, Graph)

    # Find set of nodes expected to be foci of validation results.
    expected: Set[URIRef] = {URIRef("http://example.org/kb/thing-b-1")}
    computed: Set[URIRef] = set()
    for triple in conformance_graph.triples((None, n_validation_result_property, None)):
        assert isinstance(triple[2], URIRef)
        computed.add(triple[2])
    assert expected == computed


def test_160_1() -> None:
    _test_160_template(shacl_file_1_text, SH.focusNode)


def test_160_2() -> None:
    _test_160_template(shacl_file_2_text, SH.value)



# -*- coding: utf-8 -*-
#
"""
https://github.com/RDFLib/pySHACL/issues/96
"""
from pyshacl import validate

mixed_file_text = """
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix : <urn:ex#> .

:Class0 a owl:Class .

:Class1 a owl:Class ;
    rdfs:subClassOf :Class0 .

:Class2 a owl:Class ;
    rdfs:subClassOf :Class1 .

:Class3 a owl:Class ;
    rdfs:subClassOf :Class2 .

:prop  a   owl:DatatypeProperty .

:shape a sh:NodeShape ;
    sh:targetClass :Class0 ;
    sh:property [
        sh:path :prop ;
        sh:hasValue "test" ;
        sh:minCount 1 ;
    ] .

:s2 a :Class2 ;
    :prop "fail" .

:s3 a :Class3 ;
    :prop "fail" .
"""


def test_96():
    res1 = validate(mixed_file_text, data_graph_format='turtle', shacl_graph_format='turtle', debug=True)
    conforms, _, _ = res1
    assert not conforms



import logging
import os
import sys
from functools import wraps
from io import BufferedIOBase, TextIOBase
from sys import stderr
from typing import List, Optional, Tuple, Union

from rdflib import ConjunctiveGraph, Dataset, Graph, Literal, URIRef

from pyshacl.errors import ReportableRuntimeError, ValidationFailure
from pyshacl.pytypes import GraphLike

from .consts import SH, RDF_type
from .monkey import apply_patches, rdflib_bool_patch, rdflib_bool_unpatch
from .rdfutil import load_from_source
from .rule_expand_runner import RuleExpandRunner
from .validator import Validator, assign_baked_in
from .validator_conformance import check_dash_result


def validate(
    data_graph: Union[GraphLike, BufferedIOBase, TextIOBase, str, bytes],
    *args,
    shacl_graph: Optional[Union[GraphLike, BufferedIOBase, TextIOBase, str, bytes]] = None,
    ont_graph: Optional[Union[GraphLike, BufferedIOBase, TextIOBase, str, bytes]] = None,
    advanced: Optional[bool] = False,
    inference: Optional[str] = None,
    inplace: Optional[bool] = False,
    abort_on_first: Optional[bool] = False,
    allow_infos: Optional[bool] = False,
    allow_warnings: Optional[bool] = False,
    max_validation_depth: Optional[int] = None,
    sparql_mode: Optional[bool] = False,
    focus_nodes: Optional[List[Union[str, URIRef]]] = None,
    use_shapes: Optional[List[Union[str, URIRef]]] = None,
    **kwargs,
):
    """
    :param data_graph: rdflib.Graph or file path or web url of the data to validate
    :type data_graph: rdflib.Graph | str | bytes
    :param args:
    :type args: list
    :param shacl_graph: rdflib.Graph or file path or web url of the SHACL Shapes graph to use to
    validate the data graph
    :type shacl_graph: rdflib.Graph | str | bytes
    :param ont_graph: rdflib.Graph or file path or web url of an extra ontology document to mix into the data graph
    :type ont_graph: rdflib.Graph | str | bytes
    :param advanced: Enable advanced SHACL features, default=False
    :type advanced: bool | None
    :param inference: One of "rdfs", "owlrl", "both", "none", or None
    :type inference: str | None
    :param inplace: If this is enabled, do not clone the datagraph, manipulate it in-place
    :type inplace: bool
    :param abort_on_first: Stop evaluating constraints after first violation is found
    :type abort_on_first: bool | None
    :param allow_infos: Shapes marked with severity of sh:Info will not cause result to be invalid.
    :type allow_infos: bool | None
    :param allow_warnings: Shapes marked with severity of sh:Warning or sh:Info will not cause result to be invalid.
    :type allow_warnings: bool | None
    :param max_validation_depth: The maximum number of SHACL shapes "deep" that the validator can go before reaching an "endpoint" constraint.
    :type max_validation_depth: int | None
    :param sparql_mode: Treat the DataGraph as a SPARQL endpoint, validate the graph at the SPARQL endpoint.
    :type sparql_mode: bool | None
    :param focus_nodes: A list of IRIs to validate only those nodes.
    :type focus_nodes: list | None
    :param use_shapes: A list of IRIs to use only those shapes from the SHACL ShapesGraph.
    :type use_shapes: list | None
    :param kwargs:
    :return:
    """

    do_debug = kwargs.get('debug', False)
    log = make_default_logger(name="pyshacl-validate", debug=do_debug)
    apply_patches()
    assign_baked_in()
    do_check_dash_result = kwargs.pop('check_dash_result', False)  # type: bool
    if kwargs.get('meta_shacl', False):
        to_meta_val = shacl_graph or data_graph
        conforms, v_r, v_t = meta_validate(to_meta_val, inference=inference, **kwargs)
        if not conforms:
            msg = f"SHACL File does not validate against the SHACL Shapes SHACL (MetaSHACL) file.\n{v_t}"
            log.error(msg)
            raise ReportableRuntimeError(msg)
    do_owl_imports = kwargs.pop('do_owl_imports', False)
    data_graph_format = kwargs.pop('data_graph_format', None)

    if isinstance(data_graph, (str, bytes, BufferedIOBase, TextIOBase)):
        # DataGraph is passed in as Text. It is not an rdflib.Graph
        # That means we load it into an ephemeral graph at runtime
        # that means we don't need to make a copy to prevent polluting it.
        ephemeral = True
    else:
        ephemeral = False
    use_js = kwargs.pop('js', None)
    if sparql_mode:
        if use_js:
            raise ReportableRuntimeError("Cannot use SHACL-JS in SPARQL Remote Graph Mode.")
        if inplace:
            raise ReportableRuntimeError("Cannot use inplace mode in SPARQL Remote Graph Mode.")
        if ont_graph is not None:
            raise ReportableRuntimeError("Cannot use SPARQL Remote Graph Mode with extra Ontology Graph inoculation.")
        if isinstance(data_graph, bytes):
            data_graph = data_graph.decode('utf-8')
        else:
            data_graph = data_graph
        ephemeral = False
        inplace = True
    if (
        sparql_mode
        and isinstance(data_graph, str)
        and (data_graph.lower().startswith("http:") or data_graph.lower().startswith("https:"))
    ):
        from rdflib.plugins.stores.sparqlstore import SPARQLStore

        query_endpoint: str = data_graph
        username = os.getenv("PYSHACL_SPARQL_USERNAME", "")
        method = os.getenv("PYSHACL_SPARQL_METHOD", "GET")
        auth: Optional[Tuple[str, str]]
        if username:
            password: str = os.getenv("PYSHACL_SPARQL_PASSWORD", "")
            auth = (username, password)
        else:
            auth = None
        store = SPARQLStore(query_endpoint=query_endpoint, auth=auth, method=method)
        loaded_dg = Dataset(store=store, default_union=True)
    else:
        # force no owl imports on data_graph
        loaded_dg = load_from_source(
            data_graph, rdf_format=data_graph_format, multigraph=True, do_owl_imports=False, logger=log
        )
    ont_graph_format = kwargs.pop('ont_graph_format', None)
    if ont_graph is not None:
        loaded_og = load_from_source(
            ont_graph, rdf_format=ont_graph_format, multigraph=True, do_owl_imports=do_owl_imports, logger=log
        )
    else:
        loaded_og = None
    shacl_graph_format = kwargs.pop('shacl_graph_format', None)
    if shacl_graph is not None:
        rdflib_bool_patch()
        loaded_sg = load_from_source(
            shacl_graph, rdf_format=shacl_graph_format, multigraph=True, do_owl_imports=do_owl_imports, logger=log
        )
        rdflib_bool_unpatch()
    else:
        loaded_sg = None
    iterate_rules = kwargs.pop('iterate_rules', False)
    if "abort_on_error" in kwargs:
        log.warning("Usage of abort_on_error is deprecated. Use abort_on_first instead.")
        ae = kwargs.pop("abort_on_error")
        abort_on_first = bool(abort_on_first) or bool(ae)
    validator_options_dict = {
        'debug': do_debug or False,
        'inference': inference,
        'inplace': inplace or ephemeral,
        'abort_on_first': abort_on_first,
        'allow_infos': allow_infos,
        'allow_warnings': allow_warnings,
        'advanced': advanced,
        'iterate_rules': iterate_rules,
        'use_js': use_js,
        'sparql_mode': sparql_mode,
        'logger': log,
        'focus_nodes': focus_nodes,
        'use_shapes': use_shapes,
    }
    if max_validation_depth is not None:
        validator_options_dict['max_validation_depth'] = max_validation_depth
    validator = None
    try:
        validator = Validator(
            loaded_dg,
            shacl_graph=loaded_sg,
            ont_graph=loaded_og,
            options=validator_options_dict,
        )
        conforms, report_graph, report_text = validator.run()
    except ValidationFailure as e:
        conforms = False
        report_graph = e
        report_text = "Validation Failure - {}".format(e.message)
    if do_check_dash_result and validator is not None:
        passes = check_dash_result(validator, report_graph, loaded_sg or loaded_dg)
        return passes, report_graph, report_text
    do_serialize_report_graph = kwargs.pop('serialize_report_graph', False)
    if do_serialize_report_graph and isinstance(report_graph, Graph):
        if not (isinstance(do_serialize_report_graph, str)):
            do_serialize_report_graph = 'turtle'
        report_graph = report_graph.serialize(None, encoding='utf-8', format=do_serialize_report_graph)
    return conforms, report_graph, report_text


def with_metashacl_shacl_graph_cache(f):
    # noinspection PyPep8Naming
    EMPTY = object()

    @wraps(f)
    def wrapped(*args, **kwargs):
        graph_cache = getattr(wrapped, "graph_cache", None)
        assert graph_cache is not None
        if graph_cache is EMPTY:
            import pickle

            if getattr(sys, 'frozen', False):
                # runs in a pyinstaller bundle
                here_dir = sys._MEIPASS
            else:
                here_dir = os.path.dirname(__file__)
            pickle_file = os.path.join(here_dir, "assets", "shacl-shacl.pickle")
            with open(pickle_file, 'rb') as shacl_pickle:
                u = pickle.Unpickler(shacl_pickle, fix_imports=False)
                shacl_shacl_store, identifier = u.load()
            shacl_shacl_graph = Graph(store=shacl_shacl_store, identifier=identifier)
            setattr(wrapped, "graph_cache", shacl_shacl_graph)
        return f(*args, **kwargs)

    setattr(wrapped, "graph_cache", EMPTY)
    return wrapped


@with_metashacl_shacl_graph_cache
def meta_validate(shacl_graph: Union[GraphLike, str], inference: Optional[str] = 'rdfs', **kwargs):
    shacl_shacl_graph = meta_validate.graph_cache
    shacl_graph = load_from_source(shacl_graph, rdf_format=kwargs.pop('shacl_graph_format', None), multigraph=True)
    _ = kwargs.pop('meta_shacl', None)
    return validate(shacl_graph, shacl_graph=shacl_shacl_graph, inference=inference, **kwargs)


def make_default_logger(
    name: Union[str, None] = None, debug: bool = False, clear_handlers: bool = True
) -> logging.Logger:
    log_handler = logging.StreamHandler(stderr)
    log = logging.getLogger(name)
    if clear_handlers:
        for h in log.handlers:
            log.removeHandler(h)  # pragma:no cover
    log.addHandler(log_handler)
    log.setLevel(logging.INFO if not debug else logging.DEBUG)
    log_handler.setLevel(logging.INFO if not debug else logging.DEBUG)
    return log


def shacl_rules(
    data_graph: Union[GraphLike, BufferedIOBase, TextIOBase, str, bytes],
    *args,
    shacl_graph: Optional[Union[GraphLike, BufferedIOBase, TextIOBase, str, bytes]] = None,
    ont_graph: Optional[Union[GraphLike, BufferedIOBase, TextIOBase, str, bytes]] = None,
    inference: Optional[str] = None,
    inplace: Optional[bool] = False,
    focus_nodes: Optional[List[Union[str, URIRef]]] = None,
    use_shapes: Optional[List[Union[str, URIRef]]] = None,
    **kwargs,
) -> Union[str, GraphLike]:
    """
    :param data_graph: rdflib.Graph or file path or web url of the data to validate
    :type data_graph: rdflib.Graph | str | bytes
    :param args:
    :type args: list
    :param shacl_graph: rdflib.Graph or file path or web url of the SHACL Shapes graph to use to
    validate the data graph
    :type shacl_graph: rdflib.Graph | str | bytes
    :param ont_graph: rdflib.Graph or file path or web url of an extra ontology document to mix into the data graph
    :type ont_graph: rdflib.Graph | str | bytes
    :param inference: One of "rdfs", "owlrl", "both", "none", or None
    :type inference: str | None
    :param inplace: If this is enabled, do not clone the datagraph, manipulate it in-place
    :type inplace: bool
    :param focus_nodes: A list of IRIs to validate only those nodes.
    :type focus_nodes: list | None
    :param use_shapes: A list of IRIs to use only those shapes from the SHACL ShapesGraph.
    :type use_shapes: list | None
    :param kwargs:
    :return:
    """

    do_debug = kwargs.get('debug', False)
    log = make_default_logger(name="pyshacl-rules", debug=do_debug)
    apply_patches()
    assign_baked_in()
    do_owl_imports = kwargs.pop('do_owl_imports', False)
    data_graph_format = kwargs.pop('data_graph_format', None)
    if kwargs.get('sparql_mode', None):
        raise ReportableRuntimeError("The SHACL Rules expander cannot be used in SPARQL Remote Graph Mode.")
    if isinstance(data_graph, (str, bytes, BufferedIOBase, TextIOBase)):
        # DataGraph is passed in as Text. It is not a rdflib.Graph
        # That means we load it into an ephemeral graph at runtime
        # that means we don't need to make a copy to prevent polluting it.
        ephemeral = True
    else:
        ephemeral = False
    use_js = kwargs.pop('js', None)
    # force no owl imports on data_graph
    loaded_dg = load_from_source(
        data_graph, rdf_format=data_graph_format, multigraph=True, do_owl_imports=False, logger=log
    )
    ont_graph_format = kwargs.pop('ont_graph_format', None)
    if ont_graph is not None:
        loaded_og = load_from_source(
            ont_graph, rdf_format=ont_graph_format, multigraph=True, do_owl_imports=do_owl_imports, logger=log
        )
    else:
        loaded_og = None
    shacl_graph_format = kwargs.pop('shacl_graph_format', None)
    if shacl_graph is not None:
        rdflib_bool_patch()
        loaded_sg = load_from_source(
            shacl_graph, rdf_format=shacl_graph_format, multigraph=True, do_owl_imports=do_owl_imports, logger=log
        )
        rdflib_bool_unpatch()
    else:
        loaded_sg = None
    iterate_rules = kwargs.pop('iterate_rules', False)
    runner_options_dict = {
        'debug': do_debug or False,
        'inference': inference,
        'inplace': inplace or ephemeral,
        'iterate_rules': iterate_rules,
        'use_js': use_js,
        'logger': log,
        'focus_nodes': focus_nodes,
        'use_shapes': use_shapes,
    }
    serialize_expanded_graph = kwargs.get('serialize_expanded_graph', None)
    try:
        runner = RuleExpandRunner(
            loaded_dg,
            shacl_graph=loaded_sg,
            ont_graph=loaded_og,
            options=runner_options_dict,
        )
        expanded_graph = runner.run()
    except ValidationFailure as e:
        error = "SHACL Rules Expansion Failure - {}".format(e.message)
        if serialize_expanded_graph:
            return error
        else:
            g = Graph()
            g.add((URIRef("<urn:rdflib:pyshacl:shacl-rules-error>"), RDF_type, SH.ValidationFailure))
            g.add((URIRef("<urn:rdflib:pyshacl:shacl-rules-error>"), SH.message, Literal(error)))
            return g
    if serialize_expanded_graph:
        guess_format = "trig" if isinstance(expanded_graph, (Dataset, ConjunctiveGraph)) else "turtle"
        serialize_format = kwargs.get('serialize_expanded_graph_format', guess_format)
        return expanded_graph.serialize(format=serialize_format)
    return expanded_graph



# -*- coding: utf-8 -*-
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Dict, List, Sequence, Tuple, Type, Union

from rdflib import BNode, URIRef

from pyshacl.consts import RDF_type, SH_rule, SH_SPARQLRule, SH_TripleRule
from pyshacl.errors import ReportableRuntimeError, RuleLoadError
from pyshacl.pytypes import GraphLike, RDFNode, SHACLExecutor
from pyshacl.rules.sparql import SPARQLRule
from pyshacl.rules.triple import TripleRule

if TYPE_CHECKING:
    from pyshacl.shape import Shape
    from pyshacl.shapes_graph import ShapesGraph

    from .shacl_rule import SHACLRule


def gather_rules(
    executor: SHACLExecutor,
    shacl_graph: 'ShapesGraph',
    from_shapes: Union[Sequence[Union[URIRef, BNode]], None] = None,
) -> Dict['Shape', List['SHACLRule']]:
    """
    :param executor:
    :type executor: SHACLExecutor
    :param shacl_graph:
    :type shacl_graph: ShapesGraph
    :return:
    :rtype: Dict[Shape, List[SHACLRule]]
    """
    triple_rule_nodes = set(shacl_graph.subjects(RDF_type, SH_TripleRule))
    sparql_rule_nodes = set(shacl_graph.subjects(RDF_type, SH_SPARQLRule))
    if shacl_graph.js_enabled:
        from pyshacl.extras.js.rules import JSRule, SH_JSRule

        js_rule_nodes = set(shacl_graph.subjects(RDF_type, SH_JSRule))
        use_JSRule: Union[bool, Type] = JSRule
    else:
        use_JSRule = False
        js_rule_nodes = set()
    overlaps = triple_rule_nodes.intersection(sparql_rule_nodes)
    if len(overlaps) > 0:
        raise RuleLoadError(
            "A SHACL Rule cannot be both a TripleRule and a SPARQLRule.",
            "https://www.w3.org/TR/shacl-af/#rules-syntax",
        )
    overlaps = triple_rule_nodes.intersection(js_rule_nodes)
    if len(overlaps) > 0:
        raise RuleLoadError(
            "A SHACL Rule cannot be both a TripleRule and a JSRule.",
            "https://www.w3.org/TR/shacl-af/#rules-syntax",
        )
    overlaps = sparql_rule_nodes.intersection(js_rule_nodes)
    if len(overlaps) > 0:
        raise RuleLoadError(
            "A SHACL Rule cannot be both a SPARQLRule and a JSRule.",
            "https://www.w3.org/TR/shacl-af/#rules-syntax",
        )
    used_rules = shacl_graph.subject_objects(SH_rule)
    ret_rules = defaultdict(list)
    for sub, obj in used_rules:
        if from_shapes is not None and sub not in from_shapes:
            # Skipping rule that is not in the whitelist of Shapes to use
            continue
        try:
            shape: Shape = shacl_graph.lookup_shape_from_node(sub)
        except (AttributeError, KeyError):
            raise RuleLoadError(
                "The shape that rule is attached to is not a valid SHACL Shape.",
                "https://www.w3.org/TR/shacl-af/#rules-syntax",
            )
        if obj in triple_rule_nodes:
            rule: SHACLRule = TripleRule(executor, shape, obj, iterate=executor.iterate_rules)
        elif obj in sparql_rule_nodes:
            rule = SPARQLRule(executor, shape, obj)
        elif use_JSRule and callable(use_JSRule) and obj in js_rule_nodes:
            rule = use_JSRule(executor, shape, obj)
        else:
            raise RuleLoadError(
                "when using sh:rule, the Rule must be defined as either a TripleRule or SPARQLRule.",
                "https://www.w3.org/TR/shacl-af/#rules-syntax",
            )
        ret_rules[shape].append(rule)
    return ret_rules


RULES_ITERATE_LIMIT = 100


def apply_rules(
    executor: SHACLExecutor,
    shapes_rules: Dict,
    data_graph: GraphLike,
    focus_nodes: Union[Sequence[RDFNode], None] = None,
) -> int:
    # short the shapes dict by shapes sh:order before execution
    sorted_shapes_rules: List[Tuple[Any, Any]] = sorted(shapes_rules.items(), key=lambda x: x[0].order)
    total_modified = 0
    for shape, rules in sorted_shapes_rules:
        # sort the rules by the sh:order before execution
        rules = sorted(rules, key=lambda x: x.order)
        _iterate_limit = int(RULES_ITERATE_LIMIT)
        while True:
            if _iterate_limit < 1:
                raise ReportableRuntimeError(
                    f"SHACL Shape Rule iteration exceeded iteration limit of {RULES_ITERATE_LIMIT}."
                )
            _iterate_limit -= 1
            this_modified = 0
            for r in rules:
                if r.deactivated:
                    continue
                n_modified = r.apply(data_graph, focus_nodes=focus_nodes)
                this_modified += n_modified
            if this_modified > 0:
                total_modified += this_modified
                if executor.iterate_rules:
                    continue
                else:
                    break
            break
    return total_modified



import pyshacl
from rdflib.namespace import SH


test_node_ttl = """\
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix ex: <http://example.org/> .


ex:PersonShape
  a sh:NodeShape ;
  sh:targetClass ex:Person ;
  sh:node ex:Shape1 ;
.

ex:Shape1
  a sh:NodeShape ;
  sh:node ex:Shape2 ;
  sh:property [
    a sh:PropertyShape ;
    sh:path ex:familyName ;
    sh:datatype xsd:string ;
    sh:minCount 1 ;
  ] ;
.

ex:Shape2
  a sh:NodeShape ;
  sh:property [
    a sh:PropertyShape ;
    sh:path ex:givenName ;
    sh:datatype xsd:string ;
    sh:minCount 1 ;
  ] ;
.

ex:Person1
  a ex:Person ;
.
"""


def test_node_details():
    conforms, graph, text = pyshacl.validate(test_node_ttl, data_graph_format="turtle")
    assert not conforms
    top_results = set(graph.objects(None, SH.result))
    assert len(top_results) == 1
    top_result = top_results.pop()
    details = set(graph.objects(top_result, SH.detail))
    assert len(details) == 2
    for r in details:
        details = set(graph.objects(r, SH.detail))
        if len(details) == 0:
            source_1 = r
        elif len(details) == 1:
            mid_result = r
    assert source_1
    assert mid_result
    assert (source_1, SH.sourceConstraintComponent, SH.MinCountConstraintComponent) in graph
    assert (mid_result, SH.detail / SH.sourceConstraintComponent, SH.MinCountConstraintComponent) in graph



# -*- coding: utf-8 -*-
"""
https://github.com/RDFLib/pySHACL/issues/154
"""

import rdflib

from pyshacl import validate

shacl_file = """\
@prefix ex: <urn:ex#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix schema: <http://schema.org/> .

schema:Organization
    a rdfs:Class, sh:NodeShape ;
    sh:property
    [
        sh:path schema:name ;
        sh:label "name" ;
        sh:description "The official name of the entity being described." ;
        sh:datatype xsd:string ;
        sh:minCount 1;
    ],
    [
        sh:path schema:subOrganization ;
        sh:label "contained organization" ;
        sh:description "An organization contained within a parent organization." ;
        sh:node schema:Organization ;
    ] ;
.
"""

data_file = """\
{
    "@context": "http://schema.org",
    "@id": "http://www.illinoiscourts.gov/Circuit",
    "@type":"http://schema.org/Organization",
    "name": "State of Illinois Circuit Court",
    "subOrganization":  {
        "@context": "http://schema.org/",
        "@id": "http://www.illinoiscourts.gov/Circuit#Circuit1",
        "name": "State of Illinois Circuit 1",
        "subOrganization":  {
            "@id": "http://www.illinoiscourts.gov/Circuit#Circuit1District1",
            "notname": "State of Illinois Circuit 1 District 1"
        }
    }
}
"""


def test_154() -> None:
    data_g = rdflib.Graph()
    data_g.parse(data=data_file, format="json-ld")
    shapes = rdflib.Graph()
    shapes.parse(data=shacl_file, format="turtle")
    res = validate(data_g, shacl_graph=shapes, debug=False)
    conforms, graph, string = res
    assert not conforms
    assert "Constraint Violation in NodeConstraintComponent" in string


if __name__ == "__main__":
    test_154()



from itertools import chain
from typing import TYPE_CHECKING, Dict, Optional, Union

import rdflib

from .clone import clone_blank_node, clone_dataset, clone_node
from .consts import OWL, RDF, ConjunctiveLike, GraphLike, OWL_classes, OWL_properties, RDFS_classes, RDFS_properties

if TYPE_CHECKING:
    from rdflib import BNode
    from rdflib.term import URIRef

    from .consts import RDFNode

OWLNamedIndividual = OWL.NamedIndividual


def inoculate(data_graph: rdflib.Graph, ontology: GraphLike) -> rdflib.Graph:
    """
    Copies all RDFS and OWL axioms (classes, relationship definitions, and properties)
    from the ontology graph into the data_graph.
    :param data_graph:
    :type data_graph:
    :param ontology:
    :type ontology:
    :return:
    :rtype:
    """
    copied_bnode_map: Dict[RDFNode, BNode] = {}
    copied_named_map: Dict[RDFNode, Union[BNode, RDFNode]] = {}
    ontology_ns = ontology.namespace_manager
    data_graph_ns = data_graph.namespace_manager

    if isinstance(ontology, (rdflib.ConjunctiveGraph, rdflib.Dataset)):
        # always set default_union true on the ontology DS
        ontology.default_union = True
    # Bind any missing ontology namespaces in the DataGraph NS manager.
    if ontology_ns is not data_graph_ns:
        data_graph_prefixes = {p: n for (p, n) in data_graph_ns.namespaces()}
        for p, n in ontology_ns.namespaces():
            if p not in data_graph_prefixes:
                data_graph_ns.bind(p, n)

    for ont_class in chain(RDFS_classes, OWL_classes):
        found_s = list(ontology.subjects(RDF.type, ont_class))
        for s in found_s:
            if isinstance(s, rdflib.BNode):
                if s in copied_bnode_map:
                    new_bnode = copied_bnode_map[s]
                else:
                    new_bnode = clone_blank_node(ontology, s, data_graph)
                    copied_bnode_map[s] = new_bnode
                data_graph.add((new_bnode, RDF.type, ont_class))
            else:
                if ont_class is OWLNamedIndividual:
                    if s in copied_named_map:
                        new_s = copied_named_map[s]
                    else:
                        # Whole node of NamedIndividual needs to be cloned
                        new_s = clone_node(ontology, s, data_graph, deep_clone=True)
                        copied_named_map[s] = new_s
                else:
                    # Shallow-copy this node from the ontology graph to the data graph
                    new_s = clone_node(ontology, s, data_graph, deep_clone=False)
                data_graph.add((new_s, RDF.type, ont_class))

    for ont_property in chain(RDFS_properties, OWL_properties):
        found_s_o = list(ontology.subject_objects(ont_property))
        for s2, o2 in found_s_o:
            if isinstance(s2, rdflib.BNode):
                if s2 in copied_bnode_map:
                    new_bnode = copied_bnode_map[s2]
                else:
                    new_bnode = clone_blank_node(ontology, s2, data_graph)
                    copied_bnode_map[s2] = new_bnode
                new_s = new_bnode
            else:
                new_s = s2
            if isinstance(o2, rdflib.BNode):
                if o2 in copied_bnode_map:
                    new_bnode = copied_bnode_map[o2]
                else:
                    new_bnode = clone_blank_node(ontology, o2, data_graph)
                    copied_bnode_map[o2] = new_bnode
                data_graph.add((new_s, ont_property, new_bnode))
            else:
                data_graph.add((new_s, ont_property, o2))

    # Finally add in any triples where a known NamedIndividual is the Object.
    for ni in copied_named_map.keys():
        found_s_p = list(ontology.subject_predicates(object=ni))
        for s3, p3 in found_s_p:
            if isinstance(p3, (rdflib.Literal, rdflib.BNode)):
                # predicates pointing to NamedIndividual should not be BNode or Literal
                continue
            if isinstance(s3, rdflib.BNode):
                if s3 in copied_bnode_map:
                    new_bnode = copied_bnode_map[s3]
                else:
                    new_bnode = clone_blank_node(ontology, s3, data_graph)
                    copied_bnode_map[s3] = new_bnode
                new_s = new_bnode
            else:
                new_s = s3

            data_graph.add((new_s, p3, ni))

    return data_graph


def inoculate_dataset(
    base_ds: ConjunctiveLike,
    ontology_ds: GraphLike,
    target_ds: Optional[Union[ConjunctiveLike, str]] = None,
    target_graph_identifier: Optional['URIRef'] = None,
):
    """
    Make a clone of base_ds (dataset) and add RDFS and OWL triples from ontology_ds
    :param base_ds:
    :type base_ds: rdflib.Dataset
    :param ontology_ds:
    :type ontology_ds: rdflib.Dataset|rdflib.ConjunctiveGraph|rdflib.Graph
    :param target_ds:
    :type target_ds: rdflib.Dataset|str|NoneType
    :param target_graph_identifier:
    :type target_graph_identifier: rdflib.URIRef | None
    :return: The cloned Dataset with ontology triples from ontology_ds
    :rtype: rdflib.Dataset
    """

    if target_ds is None:
        target_ds = clone_dataset(base_ds)
    elif target_ds is base_ds:
        pass
    elif target_ds == "inplace" or target_ds == "base":
        target_ds = base_ds
    elif isinstance(target_ds, str):
        raise RuntimeError("target_ds cannot be a string (unless it is 'inplace' or 'base')")

    if isinstance(target_ds, (rdflib.ConjunctiveGraph, rdflib.Dataset)):
        if not isinstance(target_ds, rdflib.Dataset):
            raise RuntimeError("Cannot inoculate ConjunctiveGraph, use Dataset instead.")
    else:
        raise RuntimeError("Cannot inoculate datasets if target_ds passed in is not a Dataset itself.")

    if target_graph_identifier:
        dest_graph = target_ds.get_context(target_graph_identifier)
    else:
        dest_graph = target_ds.default_context

    # inoculate() routine will set default_union on the ontology_ds if it is a Dataset
    inoculate(dest_graph, ontology_ds)

    return target_ds



# -*- coding: utf-8 -*-
#
from typing import Optional, Union, overload

import rdflib
from rdflib.collection import Collection
from rdflib.graph import DATASET_DEFAULT_GRAPH_ID
from rdflib.namespace import NamespaceManager

from .consts import OWL, RDF_first, RDFNode
from .pytypes import ConjunctiveLike, GraphLike

OWLsameAs = OWL.sameAs


def clone_dataset(source_ds: ConjunctiveLike, target_ds=None):
    if target_ds and not isinstance(target_ds, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
        raise RuntimeError("when cloning a dataset, the target_ds must be a conjunctiveGraph or rdflib Dataset.")
    default_union = source_ds.default_union
    if target_ds is None:
        target_ds = rdflib.Dataset(default_union=default_union)
        target_ds.namespace_manager = NamespaceManager(target_ds, 'core')
        target_ds.default_context.namespace_manager = target_ds.namespace_manager
    named_graphs = [
        (
            rdflib.Graph(source_ds.store, i, namespace_manager=source_ds.namespace_manager)  # type: ignore[arg-type]
            if not isinstance(i, rdflib.Graph)
            else i
        )
        for i in source_ds.store.contexts(None)
    ]
    if isinstance(source_ds, rdflib.Dataset) and len(named_graphs) < 1:
        named_graphs = [
            rdflib.Graph(source_ds.store, DATASET_DEFAULT_GRAPH_ID, namespace_manager=source_ds.namespace_manager)
        ]
    cloned_graphs = [
        clone_graph(ng, rdflib.Graph(target_ds.store, ng.identifier, namespace_manager=target_ds.namespace_manager))
        for ng in named_graphs
    ]
    source_graph_identifiers = [ng.identifier for ng in named_graphs]
    source_default_context_id = source_ds.default_context.identifier
    target_default_context_id = target_ds.default_context.identifier
    if source_default_context_id != target_default_context_id:
        old_target_default_context = target_ds.default_context
        old_target_default_context_id = old_target_default_context.identifier
        if isinstance(target_ds, rdflib.Dataset):
            new_target_default_context = target_ds.graph(source_default_context_id)
        else:
            new_target_default_context = target_ds.get_context(source_default_context_id)
            target_ds.store.add_graph(new_target_default_context)
        target_ds.default_context = new_target_default_context
        if old_target_default_context_id not in source_graph_identifiers:
            if isinstance(target_ds, rdflib.Dataset):
                target_ds.remove_graph(old_target_default_context)
            else:
                target_ds.store.remove_graph(old_target_default_context)
        target_default_context_id = new_target_default_context.identifier
    else:
        if isinstance(target_ds, rdflib.Dataset):
            _ = target_ds.graph(target_default_context_id)
        else:
            t_default = target_ds.get_context(target_default_context_id)
            target_ds.store.add_graph(t_default)

    for g in cloned_graphs:
        if g == target_ds.default_context or g.identifier == target_default_context_id:
            continue
        if isinstance(target_ds, rdflib.Dataset):
            _ = target_ds.graph(g)  # alias to Dataset.add_graph()
        else:
            target_ds.store.add_graph(g)
    return target_ds


def clone_graph(
    source_graph: rdflib.Graph, target_graph: Optional[rdflib.Graph] = None, identifier: Optional[str] = None
) -> rdflib.Graph:
    """
    Make a clone of the source_graph by directly copying triples from source_graph to target_graph
    :param source_graph:
    :type source_graph: rdflib.Graph
    :param target_graph:
    :type target_graph: rdflib.Graph|None
    :param identifier:
    :type identifier: str | None
    :return: The cloned graph
    :rtype: rdflib.Graph
    """
    if isinstance(source_graph, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
        return clone_dataset(source_graph, target_ds=target_graph)
    if target_graph is None:
        g = rdflib.Graph(identifier=identifier, bind_namespaces='core')
        for p, n in source_graph.namespace_manager.namespaces():
            g.namespace_manager.bind(p, n, override=True, replace=True)
    else:
        g = target_graph
        for p, n in source_graph.namespace_manager.namespaces():
            g.namespace_manager.bind(p, n, override=False, replace=False)
    for t in iter(source_graph):
        g.add(t)
    return g


def mix_datasets(
    base_ds: ConjunctiveLike, extra_ds: GraphLike, target_ds: Optional[Union[ConjunctiveLike, str]] = None
):
    """
    Make a clone of base_ds (dataset) and add in the triples from extra_ds (dataset)
    :param base_ds:
    :type base_ds: rdflib.Dataset
    :param extra_ds:
    :type extra_ds: rdflib.Dataset
    :param target_ds:
    :type target_ds: rdflib.Dataset|str|NoneType
    :return: The cloned Dataset with mixed in triples from extra_ds
    :rtype: rdflib.Dataset
    """
    default_union = base_ds.default_union
    base_named_graphs = [
        (
            rdflib.Graph(base_ds.store, i, namespace_manager=base_ds.namespace_manager)  # type: ignore[arg-type]
            if not isinstance(i, rdflib.Graph)
            else i
        )
        for i in base_ds.store.contexts(None)
    ]
    if isinstance(base_ds, rdflib.Dataset) and len(base_named_graphs) < 1:
        # rdflib.Dataset always includes the DEFAULT_GRAPH_ID named graph
        # but a conjunctive graph does not. It _could_ return no graphs.
        base_named_graphs = [
            rdflib.Graph(base_ds.store, DATASET_DEFAULT_GRAPH_ID, namespace_manager=base_ds.namespace_manager)
        ]
    if target_ds is None:
        target_ds = rdflib.Dataset(default_union=default_union)
        target_ds.namespace_manager = NamespaceManager(target_ds, 'core')
        target_ds.default_context.namespace_manager = target_ds.namespace_manager
    elif target_ds == "inplace" or target_ds == "base":
        target_ds = base_ds
    elif isinstance(target_ds, str):
        raise RuntimeError("target_ds cannot be a string (unless it is 'inplace' or 'base')")

    if isinstance(target_ds, (rdflib.ConjunctiveGraph, rdflib.Dataset)):
        if not isinstance(target_ds, rdflib.Dataset):
            raise RuntimeError("Cannot mix new graphs into a ConjunctiveGraph, use Dataset instead.")
    else:
        raise RuntimeError("Cannot mix datasets if target_ds passed in is not a Dataset itself.")

    if isinstance(extra_ds, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
        mixin_graphs = [
            (
                rdflib.Graph(extra_ds.store, i, namespace_manager=extra_ds.namespace_manager)  # type: ignore[arg-type]
                if not isinstance(i, rdflib.Graph)
                else i
            )
            for i in extra_ds.store.contexts(None)
        ]
    else:
        mixin_graphs = [extra_ds]
    if target_ds is base_ds or target_ds == "inplace" or target_ds == "base":
        target_ds = base_ds
        for mg in mixin_graphs:
            mod_named_graphs = {g.identifier: mix_graphs(g, mg, target_graph=g) for g in base_named_graphs}
    else:
        mixed_graphs = {}
        for mg in mixin_graphs:
            mod_named_graphs = {
                g.identifier: mix_graphs(
                    g,
                    mg,
                    target_graph=rdflib.Graph(
                        store=target_ds.store, namespace_manager=target_ds.namespace_manager, identifier=g.identifier
                    ),
                )
                for g in base_named_graphs
            }
            mixed_graphs.update(mod_named_graphs)

        base_graph_identifiers = [bg.identifier for bg in base_named_graphs]
        base_default_context_id = base_ds.default_context.identifier
        target_default_context_id = target_ds.default_context.identifier
        if base_default_context_id != target_default_context_id:
            old_target_default_context = target_ds.default_context
            old_target_default_context_id = old_target_default_context.identifier
            if isinstance(target_ds, rdflib.Dataset):
                new_target_default_context = target_ds.graph(base_default_context_id)
            else:
                new_target_default_context = target_ds.get_context(base_default_context_id)
                target_ds.store.add_graph(new_target_default_context)
            target_ds.default_context = new_target_default_context
            if old_target_default_context_id not in base_graph_identifiers:
                if isinstance(target_ds, rdflib.Dataset):
                    target_ds.remove_graph(old_target_default_context)
                else:
                    target_ds.store.remove_graph(old_target_default_context)
            target_default_context_id = new_target_default_context.identifier
        else:
            if isinstance(target_ds, rdflib.Dataset):
                _ = target_ds.graph(target_default_context_id)
            else:
                t_default = target_ds.get_context(target_default_context_id)
                target_ds.store.add_graph(t_default)
        for i, m in mixed_graphs.items():
            if m == target_ds.default_context or i == target_default_context_id:
                continue
            if isinstance(target_ds, rdflib.Dataset):
                _ = target_ds.graph(m)  # alias to Dataset.add_graph()
            else:
                target_ds.store.add_graph(m)
    return target_ds


def mix_graphs(base_graph: GraphLike, extra_graph: GraphLike, target_graph: Optional[Union[GraphLike, str]] = None):
    """
    Make a clone of base_graph and add in the triples from extra_graph
    :param base_graph:
    :type base_graph: rdflib.Graph
    :param extra_graph:
    :type extra_graph: rdflib.Graph
    :param target_graph:
    :type target_graph: rdflib.Graph|str|NoneType
    :return: The cloned graph with mixed in triples from extra_graph
    :rtype: rdflib.Graph
    """
    if isinstance(base_graph, (rdflib.ConjunctiveGraph, rdflib.Dataset)) and isinstance(
        target_graph, (rdflib.ConjunctiveGraph, rdflib.Dataset)
    ):
        return mix_datasets(base_graph, extra_graph, target_ds=target_graph)
    if target_graph is None:
        g = clone_graph(base_graph, target_graph=None, identifier=base_graph.identifier)
    elif target_graph == "inplace" or target_graph == "base":
        # Special case, don't clone the basegraph, just put extra straight in
        g = base_graph
    elif isinstance(target_graph, str):
        raise RuntimeError("target_graph cannot be a string (unless it is 'inplace' or 'base')")
    elif target_graph is base_graph:
        g = base_graph
    else:
        # Clone base_graph into existing target, before mixing in extra
        g = clone_graph(base_graph, target_graph=target_graph)
    g = clone_graph(extra_graph, target_graph=g)
    return g


@overload
def clone_list(
    graph: rdflib.Graph,
    lnode: rdflib.BNode,
    target_graph: rdflib.Graph,
    keepid: bool = ...,
    recursion: int = ...,
    deep_clone: bool = ...,
) -> rdflib.BNode: ...


@overload
def clone_list(
    graph: rdflib.Graph,
    lnode: rdflib.URIRef,
    target_graph: rdflib.Graph,
    keepid: bool = ...,
    recursion: int = ...,
    deep_clone: bool = ...,
) -> rdflib.URIRef: ...


def clone_list(graph, lnode, target_graph, keepid=False, recursion=0, deep_clone=False):
    # If deep_clone, copy all the contents (subjects, predicates) of a named member item
    if isinstance(lnode, rdflib.BNode):
        if keepid:
            cloned_node = rdflib.BNode(str(lnode))
        else:
            cloned_node = rdflib.BNode()
    else:
        # A list can be a NamedIndividual too
        cloned_node = rdflib.URIRef(str(lnode))
    new_list = Collection(target_graph, cloned_node)
    for item in iter(graph.items(lnode)):
        cloned_item = clone_node(graph, item, target_graph, recursion=recursion + 1, deep_clone=deep_clone)
        new_list.append(cloned_item)
    return cloned_node


def clone_blank_node(
    graph: rdflib.Graph, bnode: rdflib.BNode, target_graph: rdflib.Graph, keepid: bool = False, recursion: int = 0
) -> rdflib.BNode:
    if not isinstance(graph, rdflib.Graph):
        raise RuntimeError("clone_blank_node must take an rdflib.Graph as first parameter")
    if not isinstance(bnode, rdflib.BNode):
        raise RuntimeError("clone_blank_node must take an rdflib.BNode as second parameter")
    if keepid:
        cloned_bnode = rdflib.BNode(str(bnode))
    else:
        cloned_bnode = rdflib.BNode()
    if recursion >= 10:
        return cloned_bnode  # Cannot clone this deep

    predicates = set(graph.predicates(bnode))
    if len(predicates) < 1:
        return cloned_bnode
    if RDF_first in predicates:
        # don't increase recursion here, we're not actually going any deeper in the graph, just sideways
        return clone_list(graph, bnode, target_graph, keepid=keepid, recursion=recursion)
    for p in predicates:
        cloned_p = clone_node(graph, p, target_graph, recursion=recursion + 1)
        objs = list(graph.objects(bnode, p))
        if len(objs) < 1:
            continue
        for o in objs:
            if p is OWLsameAs and o is bnode:
                # Avoid a crazy owl:sameAs recursion with self.
                cloned_o = clone_node(graph, o, target_graph, recursion=recursion + 1, deep_clone=False)
            else:
                cloned_o = clone_node(graph, o, target_graph, recursion=recursion + 1)
            target_graph.add((cloned_bnode, cloned_p, cloned_o))
    return cloned_bnode


def clone_literal(graph: rdflib.Graph, node: rdflib.Literal, target_graph: rdflib.Graph) -> rdflib.Literal:
    lex_val_string = str(node)
    lang = node.language
    datatype = node.datatype
    new_literal = rdflib.Literal(lex_val_string, lang, datatype)
    return new_literal


def clone_node(
    graph: rdflib.Graph, node: RDFNode, target_graph: rdflib.Graph, recursion: int = 0, deep_clone: bool = False
) -> RDFNode:
    # If deepclone, when the type is URIRef, it clones _all_ node content (properties, objects)
    new_node: RDFNode
    if isinstance(node, rdflib.Literal):
        new_node = clone_literal(graph, node, target_graph)
    elif isinstance(node, rdflib.BNode):
        new_node = clone_blank_node(graph, node, target_graph, recursion=recursion + 1)
    elif isinstance(node, rdflib.URIRef):
        new_node = rdflib.URIRef(str(node))
        if deep_clone:
            # Treat it like a blankNode, and copy all the contents
            if recursion >= 10:
                return new_node  # Cannot clone this deep
            predicates = set(graph.predicates(node))
            if len(predicates) < 1:
                return new_node
            if RDF_first in predicates:
                # don't increase recursion here, we're not actually going any deeper in the graph, just sideways
                return clone_list(graph, node, target_graph, recursion=recursion)
            for p in predicates:
                cloned_p = clone_node(graph, p, target_graph, recursion=recursion + 1, deep_clone=False)
                objs = list(graph.objects(node, p))
                if len(objs) < 1:
                    continue
                for o in objs:
                    if p is OWLsameAs and o is node:
                        # Avoid a crazy owl:sameAs recursion with self.
                        cloned_o = clone_node(graph, o, target_graph, recursion=recursion + 1, deep_clone=False)
                    else:
                        cloned_o = clone_node(graph, o, target_graph, recursion=recursion + 1, deep_clone=deep_clone)
                    target_graph.add((new_node, cloned_p, cloned_o))
    else:
        raise ValueError(f"Cannot clone node of type {type(node)}")
    return new_node



# -*- coding: utf-8 -*-
#
"""
https://github.com/RDFLib/pySHACL/issues/220

"""
from pyshacl import validate
from pyshacl.rdfutil import load_from_source

data_graph1 = load_from_source("./test/issues/test_220/kb-REPRODUCTION-1.ttl")
shacl_graph1 = load_from_source("./test/issues/test_220/sh-REPRODUCTION-1.ttl")
data_graph2 = load_from_source("./test/issues/test_220/kb-REPRODUCTION-2.ttl")
shacl_graph2 = load_from_source("./test/issues/test_220/sh-REPRODUCTION-2.ttl")
ont_graph = load_from_source("./test/issues/test_220/owl-REPRODUCTION.ttl")


def test_220_1():
    conforms, g, s = validate(data_graph=data_graph1, shacl_graph=shacl_graph1, ont_graph=ont_graph, inference='none')
    assert conforms


def test_220_2():
    conforms, g, s = validate(data_graph=data_graph2, shacl_graph=shacl_graph2, ont_graph=ont_graph, inference='none')
    assert not conforms
    assert "Results (2)" in s


if __name__ == "__main__":
    test_220_1()
    test_220_2()



from pyshacl import validate
import rdflib

SCHEMA_PATH = "http://datashapes.org/schema.ttl"

data = """\
@prefix ex: <http://example.org/> .
@prefix sch: <http://schema.org/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

ex:asdgjkj a sch:CommunicateAction ;
    sch:about [ a sch:GameServer ;
            sch:playersOnline "42"^^xsd:integer ] .
"""

shacl = """\
# baseURI: http://example.org/myschema
# imports: http://datashapes.org/schema

@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix schema: <http://schema.org/> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix : <http://example.org/myschema#> .

<http://example.org/myschema>
  a owl:Ontology ;
  rdfs:comment "Dummy Schema importing from Schema.org shape"@en ;
  rdfs:label "Schema.org importer" ;
  owl:imports <http://datashapes.org/schema> .
"""


def schema_org():
    dataGraph = rdflib.Graph().parse(data=data, format='ttl')
    # print(dataGraph.serialize(format='ttl').decode('utf8'))

    shaclDS = rdflib.Dataset()
    shaclGraph = shaclDS.default_context
    shaclDS.graph(shaclGraph)
    shaclGraph.parse(data=shacl, format='ttl')

    report = validate(
        dataGraph,
        shacl_graph=shaclDS,
        abort_on_first=False,
        inference='both',
        meta_shacl=False,
        debug=False,
        advanced=True,
        do_owl_imports=True,
    )

    print(report[2])


if __name__ == "__main__":
    schema_org()



from typing import TYPE_CHECKING, Dict, Union

import rdflib

from pyshacl.consts import SH_alternativePath, SH_inversePath, SH_oneOrMorePath, SH_zeroOrMorePath, SH_zeroOrOnePath
from pyshacl.errors import ReportableRuntimeError

if TYPE_CHECKING:
    from pyshacl.shape import ShapesGraph


def shacl_path_to_sparql_path(
    shapes_graph: 'ShapesGraph', path_node, prefixes: Union[None, Dict] = None, recursion: int = 0
):
    """
    :param shapes_graph:
    :type shapes_graph: ShapesGraph
    :param path_node:
    :type path_node: rdflib.term.Node
    :param prefixes:
    :type prefixes: Union[None, Dict]
    :param recursion:
    :type recursion: int
    :returns: string
    :rtype: str
    """
    # Link: https://www.w3.org/TR/shacl/#property-paths
    if isinstance(path_node, rdflib.URIRef):
        string_uri = str(path_node)
        if prefixes is not None and len(prefixes) > 0:
            for p, ns in prefixes.items():
                if string_uri.startswith(ns):
                    string_uri = ':'.join([p, string_uri.replace(ns, '')])
                    return string_uri
        return f"<{string_uri}>"
    elif isinstance(path_node, rdflib.Literal):
        raise ReportableRuntimeError("Values of a property path cannot be a Literal.")
    # At this point, path_val _must_ be a BNode
    # TODO, the path_val BNode must be value of exactly one sh:path subject in the SG.
    if recursion >= 12:
        raise ReportableRuntimeError("Path traversal depth is too much!")
    top_level = recursion == 0
    sequence_list = list(shapes_graph.graph.items(path_node))
    if len(sequence_list) > 0:
        all_collected = []
        for s in sequence_list:
            seq1_string = shacl_path_to_sparql_path(shapes_graph, s, prefixes=prefixes, recursion=recursion + 1)
            all_collected.append(seq1_string)
        if len(all_collected) < 2:
            raise ReportableRuntimeError("List of SHACL sequence paths must have alt least two path items.")
        sequence_joined = " / ".join(all_collected)
        if top_level:
            return sequence_joined
        else:
            return f"({sequence_joined})"

    find_inverse = set(shapes_graph.objects(path_node, SH_inversePath))
    if len(find_inverse) > 0:
        inverse_path = find_inverse.pop()
        inverse_path_string = shacl_path_to_sparql_path(
            shapes_graph, inverse_path, prefixes=prefixes, recursion=recursion + 1
        )
        return f"^{inverse_path_string}"

    find_alternatives = set(shapes_graph.objects(path_node, SH_alternativePath))
    if len(find_alternatives) > 0:
        alternatives_list = find_alternatives.pop()
        all_collected = []
        for a in shapes_graph.graph.items(alternatives_list):
            alt1_string = shacl_path_to_sparql_path(shapes_graph, a, prefixes=prefixes, recursion=recursion + 1)
            all_collected.append(alt1_string)
        if len(all_collected) < 2:
            raise ReportableRuntimeError("List of SHACL alternate paths must have alt least two path items.")
        collected_joined = " | ".join(all_collected)
        if top_level:
            return collected_joined
        else:
            return f"({collected_joined})"

    find_zero_or_more = set(shapes_graph.objects(path_node, SH_zeroOrMorePath))
    if len(find_zero_or_more) > 0:
        zero_or_more_path = find_zero_or_more.pop()
        zom_path_string = shacl_path_to_sparql_path(
            shapes_graph, zero_or_more_path, prefixes=prefixes, recursion=recursion + 1
        )
        return f"{zom_path_string}*"

    find_zero_or_one = set(shapes_graph.objects(path_node, SH_zeroOrOnePath))
    if len(find_zero_or_one) > 0:
        zero_or_one_path = find_zero_or_one.pop()
        zoo_path_string = shacl_path_to_sparql_path(
            shapes_graph, zero_or_one_path, prefixes=prefixes, recursion=recursion + 1
        )
        return f"{zoo_path_string}?"

    find_one_or_more = set(shapes_graph.objects(path_node, SH_oneOrMorePath))
    if len(find_one_or_more) > 0:
        one_or_more_path = find_one_or_more.pop()
        oom_path_string = shacl_path_to_sparql_path(
            shapes_graph, one_or_more_path, prefixes=prefixes, recursion=recursion + 1
        )
        return f"{oom_path_string}+"

    raise NotImplementedError("That path method to get value nodes of property shapes is not yet implemented.")



# -*- coding: utf-8 -*-
#
from typing import Union

from rdflib import RDF, RDFS, ConjunctiveGraph, Dataset, Graph, Namespace
from rdflib.namespace import OWL
from rdflib.term import Node

RDFS_Resource = RDFS.Resource
RDF_first = RDF.first
SH = Namespace('http://www.w3.org/ns/shacl#')

ConjunctiveLike = Union[ConjunctiveGraph, Dataset]
GraphLike = Union[ConjunctiveLike, Graph]
RDFNode = Node

OWL_properties = [
    OWL.allValuesFrom,
    OWL.annotatedProperty,
    OWL.annotatedSource,
    OWL.annotatedTarget,
    OWL.assertionProperty,
    OWL.cardinality,
    OWL.complementOf,
    OWL.datatypeComplementOf,
    OWL.differentFrom,
    OWL.disjointUnionOf,
    OWL.disjointWith,
    OWL.distinctMembers,
    OWL.equivalentClass,
    OWL.equivalentProperty,
    OWL.hasKey,
    OWL.hasSelf,
    OWL.hasValue,
    OWL.intersectionOf,
    OWL.inverseOf,
    OWL.maxCardinality,
    OWL.maxQualifiedCardinality,
    OWL.members,
    OWL.minCardinality,
    OWL.minQualifiedCardinality,
    OWL.onClass,
    OWL.onDataRange,
    OWL.onDatatype,
    OWL.onProperties,
    OWL.onProperty,
    OWL.oneOf,
    OWL.propertyChainAxiom,
    OWL.propertyDisjointWith,
    OWL.qualifiedCardinality,
    OWL.sameAs,
    OWL.someValuesFrom,
    OWL.sourceIndividual,
    OWL.targetIndividual,
    OWL.targetValue,
    OWL.unionOf,
    OWL.withRestrictions,
    OWL.backwardCompatibleWith,
    OWL.deprecated,
    OWL.incompatibleWith,
    OWL.priorVersion,
    OWL.versionInfo,
    OWL.bottomDataProperty,
    OWL.topDataProperty,
    OWL.bottomObjectProperty,
    OWL.topObjectProperty,
    OWL.imports,
    OWL.versionIRI,
]
OWL_classes = [
    OWL.AllDifferent,
    OWL.AllDisjointClasses,
    OWL.AllDisjointProperties,
    OWL.Annotation,
    OWL.AnnotationProperty,
    OWL.AsymmetricProperty,
    OWL.Axiom,
    OWL.Class,
    OWL.DataRange,
    OWL.DatatypeProperty,
    OWL.DeprecatedClass,
    OWL.DeprecatedProperty,
    OWL.FunctionalProperty,
    OWL.InverseFunctionalProperty,
    OWL.IrreflexiveProperty,
    OWL.NamedIndividual,
    OWL.NegativePropertyAssertion,
    OWL.ObjectProperty,
    OWL.Ontology,
    OWL.OntologyProperty,
    OWL.ReflexiveProperty,
    OWL.Restriction,
    OWL.SymmetricProperty,
    OWL.TransitiveProperty,
]
RDFS_properties = [
    RDFS.comment,
    RDFS.domain,
    RDFS.isDefinedBy,
    RDFS.label,
    RDFS.member,
    RDFS.range,
    RDFS.seeAlso,
    RDFS.subClassOf,
    RDFS.subPropertyOf,
]
RDFS_classes = [
    RDFS.Class,
    RDFS.Container,
    RDFS.ContainerMembershipProperty,
    RDFS.Datatype,
    RDFS.Literal,
    RDFS.Resource,
]



# -*- coding: utf-8 -*-
#
"""
https://github.com/RDFLib/pySHACL/issues/26
"""
from pyshacl import validate

shacl_file_text = """
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xml: <http://www.w3.org/XML/1998/namespace> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix : <http://example.com/issue/026#> .

:PersonShape a sh:NodeShape ;
    sh:property [ rdfs:comment "Person Name" ;
            sh:datatype xsd:string ;
            sh:minCount 1 ;
            sh:maxCount 1 ;
            sh:minLength 1 ;
            sh:path foaf:name ] ;
    sh:targetClass foaf:Person .
"""

data_file_text_trig = """
@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix prov: <http://www.w3.org/ns/prov#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

<http://example.org/foaf-graph> prov:generatedAtTime "2012-04-09T00:00:00"^^xsd:dateTime .

<http://example.org/foaf-graph> {
  <http://manu.sporny.org/about#manu> a foaf:Person;
     foaf:name "Manu Sporny";
     foaf:knows <https://greggkellogg.net/foaf#me> .

  <https://greggkellogg.net/foaf#me> a foaf:Person;
     foaf:name "";
     foaf:knows <http://manu.sporny.org/about#manu> .
     
  <https://greggkellogg.net/foaf#unknown> a foaf:Person;
     foaf:name "";
     foaf:knows <https://greggkellogg.net/foaf#me> .
}
"""

data_file_text_jsonld = """
{
  "@context": {
    "generatedAt": {
      "@id": "http://www.w3.org/ns/prov#generatedAtTime",
      "@type": "http://www.w3.org/2001/XMLSchema#dateTime"
    },
    "Person": "http://xmlns.com/foaf/0.1/Person",
    "name": "http://xmlns.com/foaf/0.1/name",
    "knows": {"@id": "http://xmlns.com/foaf/0.1/knows", "@type": "@id"}
  },
  "@id": "http://example.org/foaf-graph",
  "generatedAt": "2012-04-09T00:00:00",
  "@graph": [
    {
      "@id": "http://manu.sporny.org/about#manu",
      "@type": "Person",
      "name": "Manu Sporny",
      "knows": "https://greggkellogg.net/foaf#me"
    }, {
      "@id": "https://greggkellogg.net/foaf#me",
      "@type": "Person",
      "name": "",
      "knows": "http://manu.sporny.org/about#manu"
    }, {
      "@id": "https://greggkellogg.net/foaf#unknown",
      "@type": "Person",
      "knows": "https://greggkellogg.net/foaf#me"
    }
  ]
}
"""


def test_026_trig():
    res = validate(
        data_file_text_trig,
        shacl_graph=shacl_file_text,
        data_graph_format='trig',
        shacl_graph_format='turtle',
        inference='both',
        debug=True,
    )
    conforms, graph, string = res
    assert not conforms


def test_026_jsonld():
    res = validate(
        data_file_text_jsonld,
        shacl_graph=shacl_file_text,
        data_graph_format='json-ld',
        shacl_graph_format='turtle',
        inference='both',
        debug=True,
    )
    conforms, graph, string = res
    assert not conforms


if __name__ == "__main__":
    test_026_trig()
    test_026_jsonld()



# -*- coding: utf-8 -*-

import rdflib
from packaging.version import Version
from rdflib import store

RDFLIB_VERSION = Version(rdflib.__version__)
RDFLIB_421 = Version("4.2.1")
RDFLIB_500 = Version("5.0.0")
RDFLIB_600 = Version("6.0.0")
RDFLIB_602 = Version("6.0.2")
RDFLIB_611 = Version("6.1.1")
RDFLIB_620 = Version("6.2.0")


def rdflib_bool_patch():
    from rdflib.term import _XSD_PFX, URIRef, _toPythonMapping

    rdflib.NORMALIZE_LITERALS = False
    # we want to consider only 'true' to be a valid XSD:boolean truth (ie, ignore '1')
    _toPythonMapping[URIRef(_XSD_PFX + 'boolean')] = lambda i: i.lower() == 'true'


def rdflib_bool_unpatch():
    from rdflib.term import _XSD_PFX, URIRef, _toPythonMapping

    rdflib.NORMALIZE_LITERALS = True
    if RDFLIB_500 > RDFLIB_VERSION:
        # versions before rdflib 5.0.0
        _toPythonMapping[URIRef(_XSD_PFX + 'boolean')] = lambda i: i.lower() in ['true', '1']
    else:
        # rdflib 5.0.0 and above
        from rdflib.term import _parseBoolean

        _toPythonMapping[URIRef(_XSD_PFX + 'boolean')] = _parseBoolean


def rdflib_term_ge_le_patch():
    def __le__(term, other):
        r = term.__lt__(other)
        if r:
            return r
        try:
            return term.eq(other)
        except TypeError:
            return NotImplemented

    def __ge__(term, other):
        try:
            return term.__gt__(other) or term.eq(other)
        except TypeError:
            return NotImplemented

    setattr(rdflib.term.Literal, "__ge__", __ge__)
    setattr(rdflib.term.Literal, "__le__", __le__)


def empty_iterator():
    # noinspection PyUnreachableCode
    if False:
        # noinspection PyUnreachableCode
        yield None  # type: ignore[unreachable]


def apply_patches():
    if RDFLIB_602 <= RDFLIB_VERSION < RDFLIB_611:
        # Fixes https://github.com/RDFLib/rdflib/pull/1432
        setattr(store.Store, "namespaces", empty_iterator)
    return True



# -*- coding: utf-8 -*-
#
"""
https://github.com/RDFLib/pySHACL/issues/217
"""
import sys

import rdflib

import pyshacl

shapes_data = '''\
@prefix ex: <http://example.org/ontology/> .
@prefix sh-ex: <http://example.org/shapes/> .
@prefix sh: <http://www.w3.org/ns/shacl#> .

sh-ex:ClassC-shape
    a sh:NodeShape ;
    sh:not
        [
            a sh:NodeShape ;
            sh:class ex:ClassA ;
        ] ,
        [
            a sh:NodeShape ;
            sh:class ex:ClassB ;
        ]
        ;
    sh:targetClass ex:ClassC ;
    .
'''

data_g_text = '''\
@prefix ex: <http://example.org/ontology/> .
@prefix kb: <http://example.org/kb/> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .

kb:Thing-1
    a
        ex:ClassA ,
        ex:ClassB
        ;
    rdfs:comment "This individual is consistent per OWL and should validate with SHACL."@en ;
    .

kb:Thing-2
    a
        ex:ClassA ,
        ex:ClassC
        ;
    rdfs:comment "This individual is inconsistent per OWL and should not validate with SHACL."@en ;
    .

kb:Thing-3
    a
        ex:ClassB ,
        ex:ClassC
        ;
    rdfs:comment "This individual is inconsistent per OWL and should not validate with SHACL."@en ;
    .
'''


def test_217():
    shape_g = rdflib.Graph().parse(data=shapes_data, format='turtle')
    data_g = rdflib.Graph().parse(data=data_g_text, format="turtle")
    conforms, results_graph, results_text = pyshacl.validate(
        data_g,
        shacl_graph=shape_g,
        debug=True,
        meta_shacl=False,
    )
    assert not conforms
    assert (
        "Node kb:Thing-2 must not conform to shape" in results_text
        or "Node kb:Thing-2 must not conform to any shapes" in results_text
    ) and (
        "Node kb:Thing-3 must not conform to shape" in results_text
        or "Node kb:Thing-3 must not conform to any shapes" in results_text
    )


if __name__ == "__main__":
    sys.exit(test_217())



# -*- coding: utf-8 -*-
"""
https://github.com/RDFLib/pySHACL/issues/162
"""

import rdflib

from pyshacl import validate

shacl_file = """\
@prefix ex: <urn:ex#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix schema: <http://schema.org/> .

ex:ExampleParentShape a sh:NodeShape ;
    sh:targetClass ex:test_class ;
    sh:node ex:nodeShape1,
        ex:nodeShape2 .

ex:nodeShape1 a sh:NodeShape ;
    sh:property [
        sh:path schema:name ;
        sh:class ex:test_class_1
    ] .

ex:nodeShape2 a sh:NodeShape ;
    sh:property [
        sh:path schema:subOrganization ;
        sh:class ex:test_class_1
    ] .
"""

data_file = """
@prefix ex: <urn:ex#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix schema: <http://schema.org/> .

# define ontology
ex:test_class a owl:Class .
ex:test_class_1 a owl:Class .
ex:test_class_2 a owl:Class .

# define model
ex:name_1 a ex:test_class_1 .
ex:org_1 a ex:test_class_2 . # this will cause test to fail

ex:test_entity a ex:test_class ;
    schema:name ex:name_1 ;
    schema:subOrganization ex:org_1 .

"""


def test_162() -> None:
    data_g = rdflib.Graph()
    data_g.parse(data=data_file, format="turtle")
    shapes = rdflib.Graph()
    shapes.parse(data=shacl_file, format="turtle")
    conforms, report, message = validate(data_g, shacl_graph=shapes, debug=True)
    assert not conforms
    # confirm that both nodeShapes are included in the error message
    assert "Value must conform to every Shape in ('ex:nodeShape1', 'ex:nodeShape2')" in message


if __name__ == "__main__":
    test_162()



# -*- coding: utf-8 -*-
#
import typing
from typing import Dict, List

from rdflib import XSD, Literal
from rdflib.plugins.sparql.operators import register_custom_function, unregister_custom_function
from rdflib.plugins.sparql.sparql import SPARQLError

from ..consts import SH, RDFS_comment, SH_ask, SH_parameter, SH_select
from ..errors import ConstraintLoadError, ReportableRuntimeError
from ..helper import get_query_helper_cls
from ..parameter import SHACLParameter

if typing.TYPE_CHECKING:
    from ..pytypes import GraphLike
    from ..shapes_graph import ShapesGraph


SH_returnType = SH.returnType
SH_optional = SH.optional


class SHACLFunction(object):
    __slots__ = ("sg", "node", "comments", "parameters", "rtype")

    def __init__(self, fn_node, sg: 'ShapesGraph'):
        """

        :param fn_node:
        :type fn_node: rdflib.Identifier
        :param sg:
        :type sg: ShapesGraph
        """
        super(SHACLFunction, self).__init__()
        self.node = fn_node
        self.sg = sg
        params = list(sg.objects(fn_node, SH_parameter))
        self.parameters: List[SHACLParameter] = [SHACLParameter(sg, p) for p in params]
        self.comments = set(sg.objects(fn_node, RDFS_comment))
        rtypes = list(sg.objects(fn_node, SH_returnType))
        if len(rtypes) < 1:
            self.rtype = None
        elif len(rtypes) > 1:
            raise ConstraintLoadError(
                "SHACLFunction cannot have more than one value for sh:returnType.",
                "https://www.w3.org/TR/shacl-af/#functions-example",
            )
        else:
            self.rtype = rtypes[0]

    # TODO: Maybe cache this? Its called a few times per loop
    def get_params_in_order(self):
        if len(self.parameters) < 1:
            return []
        orders = (p.param_order for p in self.parameters)
        if None not in orders:
            # sort by _param_order_ field
            params = sorted(self.parameters, key=lambda x: x.param_order)
        else:
            # sort by _localname_ of path
            params = sorted(self.parameters, key=lambda x: x.localname)
        return params

    def get_optional_map(self):
        params = self.get_params_in_order()
        return [True if p.optional else False for p in params]

    def objects(self, predicate=None):
        return self.sg.graph.objects(self.node, predicate)

    def apply(self, g):
        self.sg.add_shacl_function(self.node, self.execute, self.get_optional_map())

    def unapply(self, g):
        self.sg.remove_shacl_function(self.node, self.execute)

    def execute(self, g, *args):
        raise NotImplementedError(
            "SHACLFunction cannot be executed by itself. It needs to be a SPARQLFunction or something similar."
        )


class SPARQLFunction(SHACLFunction):
    __slots__ = ("select", "ask", "_qh")

    def __init__(self, fn_node, sg):
        super(SPARQLFunction, self).__init__(fn_node, sg)
        selects = list(self.sg.objects(self.node, SH_select))
        asks = list(self.sg.objects(self.node, SH_ask))
        num_selects = len(selects)
        num_asks = len(asks)
        if num_asks > 0 and num_selects > 0:
            raise ConstraintLoadError(
                "SPARQLFunction cannot have both sh:select and sh:ask.",
                "https://www.w3.org/TR/shacl-af/#SPARQLFunction",
            )
        elif num_asks < 1 and num_selects < 1:
            raise ConstraintLoadError(
                "SPARQLFunction must have one of either sh:select or sh:ask.",
                "https://www.w3.org/TR/shacl-af/#SPARQLFunction",
            )
        if num_selects > 1:
            raise ConstraintLoadError(
                "SPARQLFunction cannot have more than one value for sh:select.",
                "https://www.w3.org/TR/shacl-af/#SPARQLFunction",
            )
        if num_asks > 1:
            raise ConstraintLoadError(
                "SPARQLFunction cannot have more than one value for sh:ask.",
                "https://www.w3.org/TR/shacl-af/#SPARQLFunction",
            )
        if num_asks:
            self.ask = asks[0]
            if self.rtype and self.rtype != XSD.boolean:
                raise ConstraintLoadError(
                    "SPARQLFunction with sh:ask must have sh:returnType of xsd:boolean.",
                    "https://www.w3.org/TR/shacl-af/#SPARQLFunction",
                )
            else:
                self.rtype = XSD.boolean
        else:
            self.ask = None
        self.select = selects[0] if num_selects else None
        # deliberately not passing in Parameters to queryHelper here, because we can't bind them to this function
        # (this function is not a Shape, and Function Params don't get bound to it)
        SPARQLQueryHelper = get_query_helper_cls()
        query_helper = SPARQLQueryHelper(self, self.node, None, None, deactivated=False)
        query_helper.collect_prefixes()
        self._qh = query_helper

    def execute(self, g, *args):
        params = self.get_params_in_order()
        if len(args) != len(params):
            raise ReportableRuntimeError("Got incorrect number of arguments for SHACLFunction {}.".format(self.node))
        init_bindings = {}
        for i, p in enumerate(params):
            arg = args[i]
            ln = p.localname
            if arg is None and p.optional is False:
                raise ReportableRuntimeError("Got NoneType for Non-optional argument {}.".format(ln))
            init_bindings[ln] = arg
        if self.ask:
            return self.execute_ask(g, init_bindings)
        else:
            return self.execute_select(g, init_bindings)

    def execute_select(self, g: 'GraphLike', init_bindings: Dict):
        s = self._qh.apply_prefixes(self.select)
        results = g.query(s, initBindings=init_bindings)
        if results.type != "SELECT" or results.vars is None:
            raise ReportableRuntimeError("Was expecting an SELECT response from the Select query.")
        rvars = len(results.vars)
        rbindings = len(results.bindings)
        if rvars < 1 or rbindings < 1:
            return []
        rvar = results.vars[0]
        result = results.bindings[0]
        return result[rvar]

    def execute_ask(self, g: 'GraphLike', init_bindings: Dict):
        a = self._qh.apply_prefixes(self.ask)
        results = g.query(a, initBindings=init_bindings)
        if results.type != "ASK":
            raise ReportableRuntimeError("Was expecting an ASK response from the Ask query.")
        return Literal(results.askAnswer)

    def execute_from_sparql(self, e, ctx):
        if not e.expr:
            raise SPARQLError("Nothing given to SPARQLFunction.")
        params = self.get_params_in_order()
        num_params = len(params)
        if len(e.expr) > num_params:
            raise SPARQLError("Too many parameters passed to SPARQLFunction.")
        elif len(e.expr) < num_params:
            raise SPARQLError("Too few parameters passed to SPARQLFunction.")
        new_binds = ctx.ctx.initBindings.copy()
        new_binds.update(ctx.ctx.bindings)
        g = ctx.ctx.graph
        for i, var in enumerate(e.expr):
            var_val = ctx[var]
            bind_name = params[i].localname
            new_binds[bind_name] = var_val
        if self.ask:
            return self.execute_ask(g, new_binds)
        else:
            return self.execute_select(g, new_binds)

    def apply(self, g):
        super(SPARQLFunction, self).apply(g)
        register_custom_function(self.node, self.execute_from_sparql, True, True)

    def unapply(self, g):
        super(SPARQLFunction, self).unapply(g)
        unregister_custom_function(self.node, self.execute_from_sparql)



from rdflib import Graph
from pyshacl import validate

shapes_graph = '''\
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix ex: <http://example.com/ex#> .

ex:MaxLengthConstraintComponent
	a sh:ConstraintComponent ;
	sh:parameter [
		sh:path ex:maxLength ;
		sh:datatype xsd:integer ;
	] ;
	sh:validator ex:hasMaxLength .

ex:hasMaxLength
	a sh:JSValidator ;
	sh:message "Value has more than {$maxLength} characters" ;
	rdfs:comment """
		Note that $value and $maxLength are RDF nodes expressed in JavaScript Objects.
		Their string value is accessed via the .lex and .uri attributes.
		The function returns true if no violation has been found.
		""" ;
	sh:jsLibrary [ sh:jsLibraryURL "file:./test/resources/js/hasMaxLength.js"^^xsd:anyURI ] ;
	sh:jsFunctionName "hasMaxLength" .

ex:TestShape
  rdf:type sh:NodeShape ;
  rdfs:label "Test shape" ;
  sh:property [
    sh:path ex:postcode ;
    ex:maxLength 4 ;
  ] ;
  sh:targetNode ex:InvalidResource1 ;
  sh:targetNode ex:ValidResource1 ;
  .

'''

data_graph = '''\
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix ex: <http://example.com/ex#> .

ex:InvalidResource1 a rdf:Resource ;
  ex:postcode "12345" .

ex:ValidResource1 a rdf:Resource ;
  ex:postcode "1234" .
'''


def test_js_constraint_component():
    s1 = Graph().parse(data=shapes_graph, format="turtle")
    g1 = Graph().parse(data=data_graph, format="turtle")
    conforms, result_graph, result_text = validate(g1, shacl_graph=s1, advanced=True, debug=True, js=True)
    assert not conforms


if __name__ == "__main__":
    test_js_constraint_component()



# -*- coding: utf-8 -*-
"""
https://www.w3.org/TR/shacl/#core-components-count
"""

from typing import Dict, List, Optional

from rdflib.namespace import XSD
from rdflib.term import Literal, URIRef

from pyshacl.constraints.constraint_component import ConstraintComponent
from pyshacl.consts import SH
from pyshacl.errors import ConstraintLoadError
from pyshacl.pytypes import GraphLike, RDFNode, SHACLExecutor
from pyshacl.rdfutil import stringify_node
from pyshacl.shape import Shape

XSD_integer = XSD.integer
SH_minCount = SH.minCount
SH_maxCount = SH.maxCount

SH_MinCountConstraintComponent = SH.MinCountConstraintComponent
SH_MaxCountConstraintComponent = SH.MaxCountConstraintComponent


class MinCountConstraintComponent(ConstraintComponent):
    """
    sh:minCount specifies the minimum number of value nodes that satisfy the condition. If the minimum cardinality value is 0 then this constraint is always satisfied and so may be omitted.
    Link:
    https://www.w3.org/TR/shacl/#MinCountConstraintComponent
    Textual Definition:
    If the number of value nodes is less than $minCount, there is a validation result.
    """

    shacl_constraint_component = SH_MinCountConstraintComponent

    def __init__(self, shape: Shape, min_count_objects: Optional[List[RDFNode]] = None) -> None:
        super(MinCountConstraintComponent, self).__init__(shape)
        if min_count_objects is None:
            min_count = list(self.shape.objects(SH_minCount))
        else:
            min_count = min_count_objects
        if len(min_count) < 1:
            raise ConstraintLoadError(
                "MinCountConstraintComponent must have at least one sh:minCount predicate.",
                "https://www.w3.org/TR/shacl/#MinCountConstraintComponent",
            )
        if len(min_count) > 1:
            raise ConstraintLoadError(
                "MinCountConstraintComponent must have at most one sh:minCount predicate.",
                "https://www.w3.org/TR/shacl/#MinCountConstraintComponent",
            )
        if not shape.is_property_shape:
            raise ConstraintLoadError(
                "MinCountConstraintComponent can only be present on a PropertyShape, not a NodeShape.",
                "https://www.w3.org/TR/shacl/#MinCountConstraintComponent",
            )
        self.min_count = min_count[0]
        if not (isinstance(self.min_count, Literal) and self.min_count.datatype == XSD_integer):
            raise ConstraintLoadError(
                "MinCountConstraintComponent sh:minCount must be a literal with datatype xsd:integer.",
                "https://www.w3.org/TR/shacl/#MinCountConstraintComponent",
            )
        if int(self.min_count.value) < 0:
            raise ConstraintLoadError(
                "MinCountConstraintComponent sh:minCount must be an integer >= 0.",
                "https://www.w3.org/TR/shacl/#MinCountConstraintComponent",
            )

    @classmethod
    def constraint_parameters(cls) -> List[URIRef]:
        return [SH_minCount]

    @classmethod
    def constraint_name(cls) -> str:
        return "MinCountConstraintComponent"

    def make_generic_messages(self, datagraph: GraphLike, focus_node, value_node) -> List[Literal]:
        p = self.shape.path()
        if p:
            p = stringify_node(self.shape.sg.graph, p)
            try:
                focus_string = stringify_node(datagraph, focus_node)
            except (LookupError, ValueError):
                # focus node doesn't exist in the datagraph. We can deal.
                focus_string = str(focus_node)
            m = "Less than {} values on {}->{}".format(str(self.min_count.value), focus_string, p)
        else:
            m = "Less than {} values on {}".format(str(self.min_count.value), stringify_node(datagraph, focus_node))
        return [Literal(m)]

    def evaluate(
        self, executor: SHACLExecutor, target_graph: GraphLike, focus_value_nodes: Dict, _evaluation_path: List
    ):
        """
        :type executor: SHACLExecutor
        :type target_graph: rdflib.Graph
        :type focus_value_nodes: dict
        :type _evaluation_path: list
        """
        min_count = int(self.min_count.value)
        if min_count == 0:
            # MinCount of zero always passes
            return True, []
        reports = []
        non_conformant = False

        for f, value_nodes in focus_value_nodes.items():
            if not len(value_nodes) >= min_count:
                non_conformant = True
                rept = self.make_v_result(target_graph, f)
                reports.append(rept)
        return (not non_conformant), reports


class MaxCountConstraintComponent(ConstraintComponent):
    """
    sh:maxCount specifies the maximum number of value nodes that satisfy the condition.
    Link: https://www.w3.org/TR/shacl/#MaxCountConstraintComponent
    Textual Definition:
    If the number of value nodes is greater than $maxCount, there is a validation result.
    """

    shacl_constraint_component = SH_MaxCountConstraintComponent

    def __init__(self, shape: Shape, max_count_objects: Optional[List[RDFNode]] = None) -> None:
        super(MaxCountConstraintComponent, self).__init__(shape)
        if max_count_objects is None:
            max_count = list(self.shape.objects(SH_maxCount))
        else:
            max_count = max_count_objects

        if len(max_count) < 1:
            raise ConstraintLoadError(
                "MaxCountConstraintComponent must have at least one sh:maxCount predicate.",
                "https://www.w3.org/TR/shacl/#MaxCountConstraintComponent",
            )
        if len(max_count) > 1:
            raise ConstraintLoadError(
                "MaxCountConstraintComponent must have at most one sh:maxCount predicate.",
                "https://www.w3.org/TR/shacl/#MaxCountConstraintComponent",
            )
        if not shape.is_property_shape:
            raise ConstraintLoadError(
                "MaxCountConstraintComponent can only be present on a PropertyShape, not a NodeShape.",
                "https://www.w3.org/TR/shacl/#MaxCountConstraintComponent",
            )
        self.max_count = max_count[0]
        if not (isinstance(self.max_count, Literal) and self.max_count.datatype == XSD_integer):
            raise ConstraintLoadError(
                "MaxCountConstraintComponent sh:maxCount must be a literal with datatype xsd:integer.",
                "https://www.w3.org/TR/shacl/#MaxCountConstraintComponent",
            )
        if int(self.max_count.value) < 0:
            raise ConstraintLoadError(
                "MaxCountConstraintComponent sh:maxCount must be an integer >= 0.",
                "https://www.w3.org/TR/shacl/#MaxCountConstraintComponent",
            )

    @classmethod
    def constraint_parameters(cls) -> List[URIRef]:
        return [SH_maxCount]

    @classmethod
    def constraint_name(cls) -> str:
        return "MaxCountConstraintComponent"

    def make_generic_messages(self, datagraph: GraphLike, focus_node, value_node) -> List[Literal]:
        p = self.shape.path()
        if p:
            p = stringify_node(self.shape.sg.graph, p)
            m = "More than {} values on {}->{}".format(
                str(self.max_count.value), stringify_node(datagraph, focus_node), p
            )
        else:
            m = "More than {} values on {}".format(str(self.max_count.value), stringify_node(datagraph, focus_node))
        return [Literal(m)]

    def evaluate(
        self, executor: SHACLExecutor, target_graph: GraphLike, focus_value_nodes: Dict, _evaluation_path: List
    ):
        """
        :type executor: SHACLExecutor
        :type target_graph: rdflib.Graph
        :type focus_value_nodes: dict
        :type _evaluation_path: list
        """
        max_count = int(self.max_count.value)
        reports = []
        non_conformant = False

        for f, value_nodes in focus_value_nodes.items():
            if not len(value_nodes) <= max_count:
                non_conformant = True
                rept = self.make_v_result(target_graph, f)
                reports.append(rept)
        return (not non_conformant), reports



import sys

mod = sys.modules[__name__]
setattr(mod, 'SPARQLQueryHelperCls', None)


def get_query_helper_cls():
    # The SPARQLQueryHelper file is expensive to load due to regex compilation steps
    # so we do it this way so its only loaded when something actually needs to use
    # a SPARQLQueryHelper
    SPARQLQueryHelperCls = getattr(mod, 'SPARQLQueryHelperCls', None)
    if SPARQLQueryHelperCls is None:
        from .sparql_query_helper import SPARQLQueryHelper

        SPARQLQueryHelperCls = SPARQLQueryHelper
        setattr(mod, 'SPARQLQueryHelperCls', SPARQLQueryHelperCls)
    return SPARQLQueryHelperCls



# -*- coding: utf-8 -*-
"""
https://github.com/RDFLib/pySHACL/issues/190
"""

import rdflib
from pyshacl import validate
from rdflib.namespace import RDF
from rdflib.namespace import SH

shacl_file = """\
@prefix rdf:   <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix sh:    <http://www.w3.org/ns/shacl#> .
@prefix xsd:   <http://www.w3.org/2001/XMLSchema#> .
@prefix rdfs:  <http://www.w3.org/2000/01/rdf-schema#> .
@prefix ex:    <http://www.semanticweb.org/shacltest#> .
@prefix owl:   <http://www.w3.org/2002/07/owl#> .


ex:multipleSPARQLResults
    a sh:NodeShape ;
    sh:targetClass ex:Class1;
    sh:sparql [
        a sh:SPARQLConstraint ;
        sh:message "{?Class2Instance} does not have a related Class3 Instance" ;
        sh:select  '''
            PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
            PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
            PREFIX owl: <http://www.w3.org/2002/07/owl#>
            PREFIX ex: <http://www.semanticweb.org/shacltest#>
            select * where {
                ?this a ex:Class1 .
                ?Class2Instance a ex:Class2 .
                ?this ex:has_relation ?Class2Instance .
                OPTIONAL {
                    ?Class2Instance ex:has_relation ?Class3Instance .
                    ?Class3Instance a ex:Class3 .
                }
                FILTER (!bound(?Class3Instance))
            }
        ''' ;
        ] .
"""

data_file1 = """\
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix xml: <http://www.w3.org/XML/1998/namespace> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh:    <http://www.w3.org/ns/shacl#> .
@prefix ex:    <http://www.semanticweb.org/shacltest#> .

ex:has_relation rdf:type owl:ObjectProperty .

ex:Class1 rdf:type owl:Class .

ex:Class2 rdf:type owl:Class .

ex:Class3 rdf:type owl:Class .

ex:instance1Class1 rdf:type owl:NamedIndividual ,
                          ex:Class1 ;
                ex:has_relation ex:instance1Class2 ;
                ex:has_relation ex:instance2Class2 ;
                ex:has_relation ex:instance3Class2 .

ex:instance1Class2 rdf:type owl:NamedIndividual ,
                          ex:Class2 ;
                ex:has_relation ex:instance1Class3 .

ex:instance2Class2 rdf:type owl:NamedIndividual ,
                          ex:Class2 .

ex:instance3Class2 rdf:type owl:NamedIndividual ,
                          ex:Class2 .

ex:instance1Class3 rdf:type owl:NamedIndividual ,
                          ex:Class3 .
"""

data_file2 = """\
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix xml: <http://www.w3.org/XML/1998/namespace> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh:    <http://www.w3.org/ns/shacl#> .
@prefix ex:    <http://www.semanticweb.org/shacltest#> .

ex:has_relation rdf:type owl:ObjectProperty .

ex:Class1 rdf:type owl:Class .

ex:Class2 rdf:type owl:Class .

ex:Class3 rdf:type owl:Class .

ex:instance1Class1 rdf:type owl:NamedIndividual ,
                          ex:Class1 ;
                ex:has_relation ex:instance1Class2 ;
                ex:has_relation ex:instance2Class2 ;
                ex:has_relation ex:instance3Class2 .

ex:instance1Class2 rdf:type owl:NamedIndividual ,
                          ex:Class2 ;
                ex:has_relation ex:instance1Class3 .

ex:instance2Class2 rdf:type owl:NamedIndividual ,
                          ex:Class2 ;
                ex:has_relation ex:instance2Class3 .

ex:instance3Class2 rdf:type owl:NamedIndividual ,
                          ex:Class2 ;
                ex:has_relation ex:instance3Class3 .

ex:instance1Class3 rdf:type owl:NamedIndividual ,
                          ex:Class3 .

ex:instance2Class3 rdf:type owl:NamedIndividual ,
                          ex:Class3 .

ex:instance3Class3 rdf:type owl:NamedIndividual ,
                          ex:Class3 .
"""


def test_190_1() -> None:
    data_g = rdflib.Graph()
    data_g.parse(data=data_file1, format="turtle")
    shapes = rdflib.Graph()
    shapes.parse(data=shacl_file, format="turtle")

    conforms, report, message = validate(data_g, shacl_graph=shapes, debug=False)
    result_list = []
    for s, p, o in report.triples((None, RDF.type, SH.ValidationResult)):
        result_list.append(s)

    assert len(result_list) == 2


def test_190_2() -> None:
    data_g = rdflib.Graph()
    data_g.parse(data=data_file2, format="turtle")
    shapes = rdflib.Graph()
    shapes.parse(data=shacl_file, format="turtle")

    conforms, report, message = validate(data_g, shacl_graph=shapes, debug=False)
    result_list = []
    for s, p, o in report.triples((None, RDF.type, SH.ValidationResult)):
        result_list.append(s)

    assert conforms


if __name__ == "__main__":
    test_190_1()
    test_190_2()
